#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
模型输出

Authors: panxu(panxu@baidu.com)
Date:    2020/05/18 10:41:00
"""

from typing import Dict

import torch


class ModelOutputs(Dict):
    """
    模型的前向运算输出结果
    """

    def __init__(self, logits: torch.Tensor):
        """
        模型输出
        :param logits: 指模型的输出用来计算 loss
        """
        self["logits"] = logits

    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, key, value):
        self[key] = value




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
分布式训练

Authors: PanXu
Date:    2020/11/29 17:52:00
"""

from .distributed import Distributed
from .synchronized import Synchronized
from .parameter import ProcessGroupParameter, DistributedDataParallelParameter



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 sequence max label index decoder

Authors: PanXu
Date:    2020/07/05 16:42:00
"""

import torch
from easytext.tests import ASSERT

from easytext.label_decoder import SequenceMaxLabelIndexDecoder
from easytext.data import LabelVocabulary


def test_sequence_max_label_index_decoder():
    label_vocabulary = LabelVocabulary([["B-T", "B-T", "B-T", "I-T", "I-T", "O"]],
                                       padding=LabelVocabulary.PADDING)

    b_index = label_vocabulary.index("B-T")
    ASSERT.assertEqual(0, b_index)
    i_index = label_vocabulary.index("I-T")
    ASSERT.assertEqual(1, i_index)
    o_index = label_vocabulary.index("O")
    ASSERT.assertEqual(2, o_index)

    # [[O, B, I], [B, B, I], [B, I, I], [B, I, O]]
    batch_sequence_logits = torch.tensor([[[0.2, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.1, 0.7, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.1, 0.7, 0.3], [0.2, 0.3, 0.5]]],
                                         dtype=torch.float)

    expect_sequence_labels = [["O", "B-T", "I-T"],
                              ["B-T", "B-T", "I-T"],
                              ["B-T", "I-T", "I-T"],
                              ["B-T", "I-T", "O"]]

    expect = list()

    for expect_sequence_label in expect_sequence_labels:
        expect.append([label_vocabulary.index(label) for label in expect_sequence_label])

    decoder = SequenceMaxLabelIndexDecoder(label_vocabulary=label_vocabulary)

    label_indices = decoder(logits=batch_sequence_logits,
                            mask=None)

    ASSERT.assertEqual(expect, label_indices.tolist())



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
brief

Authors: panxu(panxu@baidu.com)
Date:    2020/05/13 16:26:00
"""
from typing import List

from easytext.data.tokenizer.token import Token
from easytext.data.tokenizer.tokenizer import Tokenizer


class ZhTokenizer(Tokenizer):
    """
    中文 Tokenizer
    """

    def _tokenize(self, text: str) -> List[Token]:
        """
        Tokenize
        :param text: 文本
        :return:
        """
        return [Token(t) for t in text]

    def batch_tokenize(self, batch_text: List[str]) -> List[List[Token]]:
        """
        批量 tokenize
        :param batch_text: 文本
        :return:
        """
        return [self.tokenize(t) for t in batch_text]




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
需要同步的接口

Authors: PanXu
Date:    2020/12/17 19:31:00
"""
from typing import List, Dict, Union, Tuple
from torch import Tensor
from torch.distributed import ReduceOp


class Synchronized:
    """
    需要同步数据的接口, 进行多 GPU 训练时候，训练中需要同步的数据，需要继承该接口
    """

    def to_synchronized_data(self) -> Tuple[Union[Dict[Union[str, int], Tensor], List[Tensor], Tensor], ReduceOp]:
        """
        将需要同步的数据转换成 tensor.
        :return: 需要同步的数据, 以及 同步的操作
        """

        raise NotImplementedError()

    def from_synchronized_data(self,
                               sync_data: Union[Dict[Union[str, int], Tensor], List[Tensor], Tensor],
                               reduce_op: ReduceOp) -> None:
        """
        从 tensor 数据恢复
        :param sync_data: 进行恢复的 tensor 数据
        :param reduce_op: 同步的操作，与 to_synchronized_data 的一致
        :return: None
        """
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 label 的 f1 metric

Authors: panxu(panxu@baidu.com)
Date:    2020/06/16 10:29:00
"""
import logging
import torch

from easytext.utils import log_util
from easytext.utils.json_util import json2str
from easytext.metrics import LabelF1Metric

from easytext.tests import ASSERT


# 配置log
log_util.config()


def test_label_f1_metric():
    """
    测试 label f1 metric
    """

    predictions = torch.tensor([0, 0, 1, 1, 2, 2, 3, 3, 4, 4])
    gold_labels = torch.tensor([0, 1, 1, 2, 2, 3, 3, 4, 4, 1])

    labels = [0, 1, 2, 3, 4]
    f1_metric = LabelF1Metric(labels=labels, label_vocabulary=None)

    metrics = f1_metric(prediction_labels=predictions, gold_labels=gold_labels, mask=None)

    logging.debug(json2str(metrics))

    ASSERT.assertEqual((len(labels) + 1) * 3, len(metrics))

    precision_0 = metrics[f"{LabelF1Metric.PRECISION}-0"]
    recall_0 = metrics[f"{LabelF1Metric.RECALL}-0"]
    f1_0 = metrics[f"{LabelF1Metric.F1}-0"]

    expect_precision_0 = 1. / 2.
    ASSERT.assertAlmostEqual(expect_precision_0, precision_0)
    expect_recall_0 = 1. / 1.
    ASSERT.assertAlmostEqual(expect_recall_0, recall_0)

    expect_f1_0 = 2. * expect_precision_0 * expect_recall_0 / (expect_precision_0 + expect_recall_0)
    ASSERT.assertAlmostEqual(expect_f1_0, f1_0)

    expect_precision_overall = 5. / 10.
    expect_recall_overall = 5. / 10
    precision_overall = metrics[LabelF1Metric.PRECISION_OVERALL]
    recall_overall = metrics[LabelF1Metric.RECALL_OVERALL]

    ASSERT.assertAlmostEqual(expect_precision_overall, precision_overall)
    ASSERT.assertAlmostEqual(expect_recall_overall, recall_overall)

    predictions = torch.tensor([0, 2])
    gold_labels = torch.tensor([0, 1])

    f1_metric(prediction_labels=predictions, gold_labels=gold_labels, mask=None)

    precision_0 = f1_metric.metric[f"{LabelF1Metric.PRECISION}-0"]
    recall_0 = f1_metric.metric[f"{LabelF1Metric.RECALL}-0"]
    f1_0 = f1_metric.metric[f"{LabelF1Metric.F1}-0"]

    expect_precision_0 = (1. + 1.) / (2. + 1.)
    ASSERT.assertAlmostEqual(expect_precision_0, precision_0)
    expect_recall_0 = (1. + 1.) / (1. + 1.)
    ASSERT.assertAlmostEqual(expect_recall_0, recall_0)
    expect_f1_0 = 2. * expect_precision_0 * expect_recall_0 / (expect_precision_0 + expect_recall_0)
    ASSERT.assertAlmostEqual(expect_f1_0, f1_0)


def test_label_f1_metric_with_mask():
    """
    测试 label f1 metric
    """

    predictions = torch.tensor([0, 1, 2, 3])
    gold_labels = torch.tensor([0, 0, 0, 2])
    mask = torch.tensor([1, 1, 1, 0], dtype=torch.long)

    labels = [0, 1, 2, 3]
    f1_metric = LabelF1Metric(labels=labels, label_vocabulary=None)

    metrics = f1_metric(prediction_labels=predictions, gold_labels=gold_labels, mask=mask)

    logging.debug(json2str(metrics))

    ASSERT.assertEqual((len(labels) + 1) * 3, len(metrics))

    precision_0 = metrics[f"{LabelF1Metric.PRECISION}-0"]
    recall_0 = metrics[f"{LabelF1Metric.RECALL}-0"]
    f1_0 = metrics[f"{LabelF1Metric.F1}-0"]

    expect_precision_0 = 1. / 1.
    ASSERT.assertAlmostEqual(expect_precision_0, precision_0)
    expect_recall_0 = 1. / 3.
    ASSERT.assertAlmostEqual(expect_recall_0, recall_0)

    expect_f1_0 = 2. * expect_precision_0 * expect_recall_0 / (expect_precision_0 + expect_recall_0)
    ASSERT.assertAlmostEqual(expect_f1_0, f1_0)

    expect_precision_overall = 1. / 3.
    expect_recall_overall = 1. / 3.
    precision_overall = metrics[LabelF1Metric.PRECISION_OVERALL]
    recall_overall = metrics[LabelF1Metric.RECALL_OVERALL]

    ASSERT.assertAlmostEqual(expect_precision_overall, precision_overall)
    ASSERT.assertAlmostEqual(expect_recall_overall, recall_overall)




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2021 PanXu, Inc. All Rights Reserved
#
"""
tensor 的工具

Authors: PanXu
Date:    2021/02/15 12:32:00
"""

import torch


def is_tensor_equal(tensor1: torch.Tensor, tensor2: torch.Tensor, epsilon: float = 1e-9) -> bool:
    """
    判断两个 tensor 是否相等, 如果两个 tensor 的 size 不同，直接会抛出异常
    :param tensor1: 第一个
    :param tensor2:
    :param epsilon:
    :return: True: 相等; False: 不相等
    """

    assert tensor1.size() == tensor2.size(), f"tensor1 size: {tensor1.size()} 与 tensor2 size: {tensor2.size()} 不匹配"

    if tensor1.dtype == torch.long and tensor2.dtype == torch.long:
        # 都是整数的时候，需要将 epsilon 设置为 0
        epsilon = 0

    delta = tensor1 - tensor2

    delta_mask = delta.abs() > epsilon

    return delta_mask.sum() == 0



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2021 PanXu, Inc. All Rights Reserved
#
"""
测试 tensor util

Authors: PanXu
Date:    2021/02/15 12:40:00
"""

import torch

from easytext.utils.nn import tensor_util
from easytext.tests import ASSERT


def test_is_tensor_equal():
    """
    测试两个 tensor 是否相等
    :return:
    """

    x = torch.tensor([1, 2, 3])
    y = torch.tensor([1, 2, 3])

    equal = tensor_util.is_tensor_equal(tensor1=x, tensor2=y, epsilon=0)

    ASSERT.assertTrue(equal)

    x = torch.tensor([1, 2, 3])
    y = torch.tensor([2, 2, 3])

    equal = tensor_util.is_tensor_equal(tensor1=x, tensor2=y, epsilon=0)

    ASSERT.assertFalse(equal)

    x = torch.tensor([1.0001, 2.0001, 3.0001])
    y = torch.tensor([1., 2., 3.])

    equal = tensor_util.is_tensor_equal(tensor1=x, tensor2=y, epsilon=1e-3)

    ASSERT.assertTrue(equal)

    equal = tensor_util.is_tensor_equal(tensor1=x, tensor2=y, epsilon=1e-4)

    ASSERT.assertFalse(equal)





#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
基础的 trainer callback

Authors: PanXu
Date:    2020/10/19 11:00:00
"""

from torch.utils.tensorboard import SummaryWriter

from easytext.trainer import Record
from easytext.trainer.trainer_callback import TimeConsumeCallback
from easytext.trainer.trainer_callback import LossCallback
from easytext.trainer.trainer_callback import MetricCallback
from easytext.trainer.trainer_callback import TrainerCallbackComposite


class BasicTrainerCallbackComposite(TrainerCallbackComposite):
    """
    basic trainer callback
    """

    def __init__(self, tensorboard_log_dir: str):
        """
        初始化
        :param tensorboard_log_dir: tensorboard 的 log dir
        """
        super().__init__()
        self._summary_writer = SummaryWriter(log_dir=tensorboard_log_dir)

        self._callbacks.add(TimeConsumeCallback(tensorboard_summary_writer=self._summary_writer))
        self._callbacks.add(LossCallback(tensorboard_summary_writer=self._summary_writer))
        self._callbacks.add(MetricCallback(tensorboard_summary_writer=self._summary_writer))

    def on_training_complete(self, trainer: "Trainer", record: Record) -> None:
        self._summary_writer.close()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
测试 Vocabulary

Authors: panxu(panxu@baidu.com)
Date:    2020/05/20 14:43:00
"""
import os
from unittest import TestCase

from easytext.data import Vocabulary, LabelVocabulary
from easytext.tests import ROOT_PATH, ASSERT


def test_vocabulary():
    """

    :return:
    """

    batch_tokens = [["我", "和", "你"], ["在", "我"]]
    vocabulary = Vocabulary(batch_tokens,
                            padding="",
                            unk="",
                            special_first=True,
                            min_frequency=1,
                            max_size=None)

    ASSERT.assertEqual(vocabulary.size, 4)

    ASSERT.assertTrue(not vocabulary.padding)
    ASSERT.assertTrue(not vocabulary.unk)

    ASSERT.assertEqual(vocabulary.index("我"), 0)
    ASSERT.assertEqual(vocabulary.index("和"), 1)


def test_vocabulary_speical_first():
    """
    测试 vocabulary speical first
    :return:
    """
    batch_tokens = [["我", "和", "你"], ["在", "我"]]
    vocabulary = Vocabulary(batch_tokens,
                            padding=Vocabulary.PADDING,
                            unk=Vocabulary.UNK,
                            special_first=True,
                            min_frequency=1,
                            max_size=None)

    ASSERT.assertEqual(vocabulary.size, 6)

    ASSERT.assertEqual(vocabulary.padding, vocabulary.PADDING)
    ASSERT.assertEqual(vocabulary.unk, vocabulary.UNK)
    ASSERT.assertEqual(vocabulary.index(vocabulary.padding), 0)
    ASSERT.assertEqual(vocabulary.index(vocabulary.unk), 1)


def test_other_speical_tokens():
    """
    测试 vocabulary speical first
    :return:
    """
    batch_tokens = [["我", "和", "你"], ["在", "我"]]
    vocabulary = Vocabulary(batch_tokens,
                            padding=Vocabulary.PADDING,
                            unk=Vocabulary.UNK,
                            special_first=True,
                            other_special_tokens=["<Start>", "<End>"],
                            min_frequency=1,
                            max_size=None)

    ASSERT.assertEqual(vocabulary.size, 8)

    ASSERT.assertEqual(vocabulary.padding, vocabulary.PADDING)
    ASSERT.assertEqual(vocabulary.unk, vocabulary.UNK)
    ASSERT.assertEqual(vocabulary.index(vocabulary.padding), 0)
    ASSERT.assertEqual(vocabulary.index(vocabulary.unk), 1)
    ASSERT.assertEqual(vocabulary.index("<Start>"), 2)
    ASSERT.assertEqual(vocabulary.index("<End>"), 3)


def test_speical_last():
    batch_tokens = [["我", "和", "你"], ["在", "我"]]
    vocabulary = Vocabulary(batch_tokens,
                            padding=Vocabulary.PADDING,
                            unk=Vocabulary.UNK,
                            special_first=False,
                            other_special_tokens=["<Start>", "<End>"],
                            min_frequency=1,
                            max_size=None)

    ASSERT.assertEqual(vocabulary.size, 8)

    ASSERT.assertEqual(vocabulary.padding, vocabulary.PADDING)
    ASSERT.assertEqual(vocabulary.unk, vocabulary.UNK)
    ASSERT.assertEqual(vocabulary.index(vocabulary.padding), 3 + 1)
    ASSERT.assertEqual(vocabulary.index(vocabulary.unk), 3 + 2)
    ASSERT.assertEqual(vocabulary.index("<Start>"), 3 + 3)
    ASSERT.assertEqual(vocabulary.index("<End>"), 3 + 4)


def test_label_vocabulary():
    """
    测试 label vocabulary
    :return:
    """
    vocabulary = LabelVocabulary([["A", "B", "C"], ["D", "E"]], padding="")
    ASSERT.assertEqual(vocabulary.size, 5)

    vocabulary = LabelVocabulary([["A", "B", "C"], ["D", "E"]], padding=LabelVocabulary.PADDING)
    ASSERT.assertEqual(vocabulary.size, 6)
    ASSERT.assertEqual(vocabulary.label_size, 5)

    ASSERT.assertEqual(vocabulary.index(vocabulary.padding), 5)

    for index, w in enumerate(["A", "B", "C", "D", "E"]):
        ASSERT.assertEqual(vocabulary.index(w), index)


def test_save_and_load():
    """
    测试存储和载入 vocabulary
    :return:
    """
    batch_tokens = [["我", "和", "你"], ["在", "我"], ["newline\nnewline"]]
    vocabulary = Vocabulary(batch_tokens,
                            padding=Vocabulary.PADDING,
                            unk=Vocabulary.UNK,
                            special_first=True,
                            other_special_tokens=["<Start>", "<End>"],
                            min_frequency=1,
                            max_size=None)

    ASSERT.assertEqual(vocabulary.size, 9)

    ASSERT.assertEqual(vocabulary.padding, vocabulary.PADDING)
    ASSERT.assertEqual(vocabulary.unk, vocabulary.UNK)
    ASSERT.assertEqual(vocabulary.index(vocabulary.padding), 0)
    ASSERT.assertEqual(vocabulary.index(vocabulary.unk), 1)
    ASSERT.assertEqual(vocabulary.index("<Start>"), 2)
    ASSERT.assertEqual(vocabulary.index("<End>"), 3)
    ASSERT.assertEqual(vocabulary.index("我"), 4)
    ASSERT.assertEqual(vocabulary.index("newline\nnewline"), 8)
    ASSERT.assertEqual(vocabulary.index("哈哈"), vocabulary.index(vocabulary.unk))

    vocab_dir = os.path.join(ROOT_PATH, "data/easytext/tests")

    if not os.path.isdir(vocab_dir):
        os.makedirs(vocab_dir, exist_ok=True)

    vocabulary.save_to_file(vocab_dir)

    loaded_vocab = Vocabulary.from_file(directory=vocab_dir)

    ASSERT.assertEqual(vocabulary.size, 9)

    ASSERT.assertEqual(loaded_vocab.padding, vocabulary.PADDING)
    ASSERT.assertEqual(loaded_vocab.unk, vocabulary.UNK)
    ASSERT.assertEqual(loaded_vocab.index(vocabulary.padding), 0)
    ASSERT.assertEqual(loaded_vocab.index(vocabulary.unk), 1)
    ASSERT.assertEqual(loaded_vocab.index("<Start>"), 2)
    ASSERT.assertEqual(loaded_vocab.index("<End>"), 3)
    ASSERT.assertEqual(vocabulary.index("我"), 4)
    ASSERT.assertEqual(vocabulary.index("newline\nnewline"), 8)
    ASSERT.assertEqual(vocabulary.index("哈哈"), vocabulary.index(vocabulary.unk))









#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
trainer callback

Authors: PanXu
Date:    2020/10/13 09:42:00
"""

from .trainer_callback import TrainerCallback
from .distributed_callback import DistributedCallback
from .time_consume_callback import TimeConsumeCallback
from .loss_callback import LossCallback
from .metric_callback import MetricCallback
from .trainer_callback_composite import TrainerCallbackComposite
from .basic_trainer_callback_compostie import BasicTrainerCallbackComposite



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2021 PanXu, Inc. All Rights Reserved
#
"""
接口: 将数据转换成一个 instance, 也就是一个训练的样本

Authors: PanXu
Date:    2021/08/19 08:54:00
"""


class InstanceFactory:
    """
    接口: 将数据转换成一个 instance, 也就是一个训练的样本
    训练时: 在构建数据集的时候，会构建 instance
    预测时: 在预测的时候，也会使用同样的方法来构建 instance
    这样做的收益在于，保证了训练和预测时候使用了同样的数据处理方式。
    """

    def create_instance(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
brief

Authors: PanXu
Date:    2020/09/13 15:29:00
"""

from easytext.data.tokenizer import ZhTokenizer

def test_zh_tokenizer():

    tokenizer = ZhTokenizer(is_remove_invalidate_char=False)





#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
brief

Authors: panxu(panxu@baidu.com)
Date:    2020/05/19 18:36:00
"""
from typing import Dict, List, Union
import torch
from collections import defaultdict

from easytext.data import LabelVocabulary
from easytext.utils import bio as BIO
from .metric import Metric
from .f1_metric import F1Metric


class SpanF1Metric(F1Metric):
    """
    计算基于 BIO, BIOUL 形式的 f1
    """

    def __init__(self,
                 label_vocabulary: LabelVocabulary) -> None:
        """
        初始化
        :param label_vocabulary: label 的 vocabulary
        """
        labels = set()

        # 从 B-Label, I-Label, 中获取 Label
        for index in range(label_vocabulary.label_size):
            bio_label: str = label_vocabulary.token(index)

            if bio_label == "O":
                continue

            label = bio_label.split("-")[1]
            labels.add(label)

        labels = [_ for _ in labels]

        super().__init__(labels=labels)
        self.label_vocabulary = label_vocabulary

    @property
    def schema(self):
        return "BIO"

    def __call__(self,
                 prediction_labels: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: torch.Tensor) -> Dict:
        """
        计算 metric. 返回的是 F1 字典:

        {"precision_[tag]": [value],
         "recall_[tag]" : [value],
         "f1-measure_[tag]": [value],
         "precision-overall": [value],
         "recall-overall": [value],
         "f1-measure-overall": [value]}

         其中的 [tag] 是 span 的 tag, 也就是 "B-[tag]" 中的 "[tag]"

        :param prediction_labels: 预测的结果, shape: (B, SeqLen)
        :param gold_labels: 实际的结果, shape: (B, SeqLen)
        :param mask: 对 predictions 和 gold label 的 mask, shape: (B, SeqLen)
        :return: 当前的 metric 计算字典结果.
        """

        if prediction_labels.dim() != 2:
            raise RuntimeError(f"prediction_labels shape 应该是: (B, SeqLen), 现在是:{prediction_labels.size()}")
        if gold_labels.dim() != 2:
            raise RuntimeError(f"gold_labels shape 应该是: (B, SeqLen), 现在是:{gold_labels.size()}")

        if mask is not None:
            if mask.dim() != 2:
                raise RuntimeError(f"mask shape 应该是: (B, SeqLen), 现在是:{mask.size()}")

        # 转换到 cpu 进行计算
        prediction_labels, gold_labels = prediction_labels.detach().cpu(), gold_labels.detach().cpu()

        if mask is not None:
            mask = mask.detach().cpu()
        else:
            mask = torch.ones(size=(prediction_labels.size(0), prediction_labels.size(1)),
                              dtype=torch.long).cpu()

        assert prediction_labels.size() == gold_labels.size(), \
            f"prediction_labels.size: {prediction_labels.size()} 与 gold_labels.size: {gold_labels.size()} 不匹配!"

        assert prediction_labels.size() == mask.size(), \
            f"prediction_labels.size: {prediction_labels.size()} 与 mask.size: {mask.size()} 不匹配!"

        bool_mask = (mask != 0)

        num_classes = self.label_vocabulary.label_size

        if (torch.masked_select(gold_labels, bool_mask) >= num_classes).any():
            raise RuntimeError(f"gold_labels 中存在比 num_classes 大的数值")

        # 将预测的结果 decode 成 span list
        prediction_spans_list = BIO.decode_label_index_to_span(batch_sequence_label_index=prediction_labels,
                                                               mask=mask,
                                                               vocabulary=self.label_vocabulary)

        # 将gold label index decode 成 span  list
        gold_spans_list = BIO.decode_label_index_to_span(batch_sequence_label_index=gold_labels,
                                                         mask=mask,
                                                         vocabulary=self.label_vocabulary)

        # 预测的 每个 label 的 span 数量字典
        num_prediction = defaultdict(int)

        # golden 每一个 label 的 span
        num_golden = defaultdict(int)

        # 当前 batch 下的 true_positives
        true_positives = defaultdict(int)
        false_positives = defaultdict(int)
        false_negatives = defaultdict(int)

        for prediction_spans, gold_spans in zip(prediction_spans_list, gold_spans_list):
            intersection = BIO.span_intersection(span_list1=prediction_spans,
                                                 span_list2=gold_spans)

            for span in intersection:
                # self._true_positives[span["label"]] += 1
                true_positives[span["label"]] += 1

            for span in prediction_spans:
                num_prediction[span["label"]] += 1

            for span in gold_spans:
                num_golden[span["label"]] += 1

        for label, num in num_prediction.items():
            false_positives[label] = num - true_positives[label]

        for label, num in num_golden.items():
            false_negatives[label] = num - true_positives[label]

        for k, v in true_positives.items():
            self._true_positives[k] += v

        for k, v in false_positives.items():
            self._false_positives[k] += v

        for k, v in false_negatives.items():
            self._false_negatives[k] += v

        return self._metric(true_positives=true_positives,
                            false_positives=false_positives,
                            false_negatives=false_negatives)




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
测试 span f1 measure

Authors: panxu(panxu@baidu.com)
Date:    2020/05/21 19:21:00
"""
import json
import torch

from easytext.data.vocabulary import LabelVocabulary
from easytext.tests import ASSERT
from easytext.metrics import ModelTargetMetric
from easytext.metrics.span_f1_metric import SpanF1Metric

VOCAB = LabelVocabulary([["B-T", "B-T", "B-T", "I-T", "I-T", "O"]],
                                 padding=LabelVocabulary.PADDING)

b_index = VOCAB.index("B-T")
ASSERT.assertEqual(0, b_index)
i_index = VOCAB.index("I-T")
ASSERT.assertEqual(1, i_index)
o_index = VOCAB.index("O")
ASSERT.assertEqual(2, o_index)


def test_span_f1_measure():

    # [[O, B, I], [B, B, I], [B, I, I], [B, I, O]]
    batch_sequence_logits = torch.tensor([[[0.2, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.1, 0.7, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.1, 0.7, 0.3], [0.2, 0.3, 0.5]]],
                                         dtype=torch.float)

    batch_sequence_labels = [["O", "B-T", "I-T"],
                             ["B-T", "B-T", "I-T"],
                             ["B-T", "I-T", "I-T"],
                             ["B-T", "I-T", "O"]]
    sequence_label_indices = list()

    for sequence_label in batch_sequence_labels:
        sequence_label_indices.append([VOCAB.index(label) for label in sequence_label])

    sequence_label_indices = torch.tensor(sequence_label_indices, dtype=torch.long)

    gold = torch.tensor([
        [2, 0, 1],
        [0, 0, 1],
        [0, 1, 1],
        [0, 1, 2]
    ])

    f1 = SpanF1Metric(label_vocabulary=VOCAB)

    f1(prediction_labels=sequence_label_indices, gold_labels=gold, mask=None)

    metrics = f1.metric

    print(f"metrics: {json.dumps(metrics)}")

    expect = {f"{SpanF1Metric.PRECISION}-T": 1., f"{SpanF1Metric.RECALL}-T": 1., f"{SpanF1Metric.F1}-T": 1.,
              SpanF1Metric.PRECISION_OVERALL: 1., SpanF1Metric.RECALL_OVERALL: 1., SpanF1Metric.F1_OVERALL: 1.}

    for key, _ in expect.items():
        ASSERT.assertAlmostEqual(expect[key], metrics[key])


def test_span_f1_measure_with_mask():

    # [[O, B, I], [B, B, I], [B, I, I], [B, I, O]]
    batch_sequence_logits = torch.tensor([[[0.2, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.1, 0.7, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.1, 0.7, 0.3], [0.2, 0.3, 0.5]]],
                                         dtype=torch.float)

    batch_sequence_labels = [["O", "B-T", "I-T"],
                             ["B-T", "B-T", "I-T"],
                             ["B-T", "I-T", "I-T"],
                             ["B-T", "I-T", "O"]]
    sequence_label_indices = list()

    for sequence_label in batch_sequence_labels:
        sequence_label_indices.append([VOCAB.index(label) for label in sequence_label])

    sequence_label_indices = torch.tensor(sequence_label_indices, dtype=torch.long)

    gold = torch.tensor([
        [2, 0, 1],
        [0, 0, 1],
        [0, 1, 1],
        [0, 1, 2]
    ])

    f1 = SpanF1Metric(label_vocabulary=VOCAB)

    mask = torch.tensor([
        [1, 1, 0],
        [1, 1, 1],
        [1, 0, 0],
        [1, 1, 1]
    ], dtype=torch.long)

    f1(prediction_labels=sequence_label_indices, gold_labels=gold, mask=mask)

    metrics = f1.metric

    print(f"metrics: {json.dumps(metrics)}")

    expect = {f"{SpanF1Metric.PRECISION}-T": 1., f"{SpanF1Metric.RECALL}-T": 1., f"{SpanF1Metric.F1}-T": 1.,
              f"{SpanF1Metric.PRECISION_OVERALL}": 1., f"{SpanF1Metric.RECALL_OVERALL}": 1., f"{SpanF1Metric.F1_OVERALL}": 1.}

    for key, _ in expect.items():
        ASSERT.assertAlmostEqual(expect[key], metrics[key])


def test_span_f1_measure_part_match():

    # [[O, B, I], [B, B, I], [B, I, I], [B, I, O]]
    batch_sequence_logits = torch.tensor([[[0.2, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.1, 0.7, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.1, 0.7, 0.3], [0.2, 0.3, 0.5]]],
                                         dtype=torch.float)

    batch_sequence_labels = [["O", "B-T", "I-T"],
                             ["B-T", "B-T", "I-T"],
                             ["B-T", "I-T", "I-T"],
                             ["B-T", "I-T", "O"]]
    sequence_label_indices = list()

    for sequence_label in batch_sequence_labels:
        sequence_label_indices.append([VOCAB.index(label) for label in sequence_label])

    sequence_label_indices = torch.tensor(sequence_label_indices, dtype=torch.long)

    gold = torch.tensor([
        [2, 0, 0],
        [0, 0, 1],
        [0, 1, 1],
        [0, 1, 2]
    ])

    f1 = SpanF1Metric(label_vocabulary=VOCAB)

    f1(prediction_labels=sequence_label_indices, gold_labels=gold, mask=None)

    metrics = f1.metric

    print(f"metrics: {json.dumps(metrics)}")

    expect = {f"{SpanF1Metric.PRECISION}-T": 0.8, f"{SpanF1Metric.RECALL}-T": 2/3,
              f"{SpanF1Metric.PRECISION_OVERALL}": 0.8, f"{SpanF1Metric.RECALL_OVERALL}": 2/3}

    for key, _ in expect.items():
        ASSERT.assertAlmostEqual(expect[key], metrics[key])




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
learning rate scheduler factory

Authors: panxu(panxu@baidu.com)
Date:    2020/06/03 09:58:00
"""

from torch.optim import Optimizer
from torch.optim import lr_scheduler
from easytext.model import Model


class LRSchedulerFactory:

    def create(self, optimizer: Optimizer, model: Model) -> "lr_scheduler":
        """
        创建 Learning Rate Scheduler
        :param optimizer: 优化器
        :param model: 模型
        :return: lr scheduler 对象
        """
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
序列标注用的 bioul

Authors: panxu(panxu@baidu.com)
Date:    2020/05/13 17:07:00
"""

import logging
from typing import List, Dict, Tuple

import torch
from easytext.data import LabelVocabulary


def fill(sequence_label: List[str], begin_index: int, end_index: int, tag: str) -> "None":
    """
    对 sequence label 填充 B I O
    :param sequence_label: 需要填充的序列
    :param begin_index: 开始位置
    :param end_index: 结束位置 [begin_index, end_index)
    :param tag: 对 sequence_label 中填充的 "B-tag", "I-tag"
    :return:
    """
    if len(sequence_label) == 0:
        raise RuntimeError("sequence_label size is 0")

    sequence_label[begin_index:end_index] = [f"I-{tag}"] * (end_index - begin_index)
    sequence_label[begin_index] = f"B-{tag}"


def decode_one_sequence_logits_to_label(sequence_logits: torch.Tensor,
                                        vocabulary: LabelVocabulary) -> Tuple[List[str], List[int]]:
    """
    对 输出 sequence logits 进行解码, 是仅仅一个 sequence 进行解码，而不是 batch sequence 进行解码。
    batch sequence 解码需要进行循环
    :param sequence_logits: shape: (seq_len, label_num),
    是 mask 之后的有效 sequence，而不是包含 mask 的 sequecne logits.
    :return: sequence label, B, I, O 的list 以及 label 对应的 index list
    """

    if len(sequence_logits.shape) != 2:
        raise RuntimeError(f"sequence_logits shape 是 (seq_len, label_num)， 现在是 {sequence_logits.shape}")

    idel_state, span_state = 0, 1

    sequence_length = sequence_logits.size(0)

    state = idel_state

    # 按权重进行排序 indices shape: (seq_len, label_num)
    sorted_sequence_indices = torch.argsort(sequence_logits, dim=-1, descending=True)

    sequence_label = list()
    sequence_label_indices = list()

    for i in range(sequence_length):

        indices = sorted_sequence_indices[i, :].tolist()

        if state == idel_state:

            # 循环寻找，直到找到一个合理的标签
            for index in indices:

                label = vocabulary.token(index)

                if label[0] == "O":

                    sequence_label.append(label)
                    sequence_label_indices.append(index)
                    state = idel_state
                    break
                elif label[0] == "B":
                    sequence_label.append(label)
                    sequence_label_indices.append(index)
                    state = span_state
                    break
                else:
                    # 其他情况 "I" 这是不合理的，所以这个逻辑是找到一个合理的标签
                    pass

        elif state == span_state:
            for index in indices:

                label = vocabulary.token(index)

                if label[0] == "B":
                    sequence_label.append(label)
                    sequence_label_indices.append(index)
                    state = span_state
                    break
                elif label[0] == "O":
                    sequence_label.append(label)
                    sequence_label_indices.append(index)
                    state = idel_state
                    break
                elif label[0] == "I":
                    sequence_label.append(label)
                    sequence_label_indices.append(index)
                    state = span_state
                    break
                else:
                    raise RuntimeError(f"{label} 不符合 BIO 格式")
        else:
            raise RuntimeError(f"state is error: {state}")

    return sequence_label, sequence_label_indices


def decode_one_sequence_label_to_span(sequence_label: List[str]) -> List[Dict]:
    """
    对 BIO 序列进行解码成 List. 例如:

    ["B-Per", "I-Per", "O", "B-Loc"] ->
    [ {"label": "Per", "begin": 0, "end": 2},
      {"label": "Loc", "begin": 3, "end": 4} ]
    :param sequence_label: BIO 序列。
    :return: 解码好的字典列表
    """
    idel_state, span_state = 0, 1

    spans = list()
    begin = None
    tag = None

    state = idel_state
    for i, label in enumerate(sequence_label):

        if state == idel_state:
            if label[0] == "B":
                begin = i
                tag = label[2:]
                state = span_state
            elif label[0] == "O":
                pass
            elif label[0] == "I":
                logging.warning(f"{sequence_label} 有不满足 BIO 格式的问题")
            else:
                raise RuntimeError(f"{label} schema 不符合 BIO")

        elif state == span_state:
            if label[0] == "B":
                span = {"label": tag,
                        "begin": begin,
                        "end": i}
                spans.append(span)
                begin = i
                tag = label[2:]
                state = span_state
            elif label[0] == "O":
                span = {"label": tag,
                        "begin": begin,
                        "end": i}
                spans.append(span)
                begin = None
                tag = None
                state = idel_state
            elif label[0] == "I":
                state = span_state
            else:
                raise RuntimeError(f"{label} schema 不符合 BIO")
        else:
            raise RuntimeError(f"{state} 错误，应该是 在 [{idel_state}, {span_state}] ")

    if state == span_state:
        span = {"label": tag,
                "begin": begin,
                "end": len(sequence_label)}
        spans.append(span)

    return spans


def decode(batch_sequence_logits: torch.Tensor,
           mask: torch.LongTensor,
           vocabulary: LabelVocabulary) -> List[List[Dict]]:
    """
    对 sequence logits 进行 BIO 解码，得到 span.

    例如:
    batch_sequence_logits:
    [[[0.1, 0.7, 0.2], [0.2, 0.2, 0.6], [0.5, 0.2, 0.3]],
     [[0.1, 0.7, 0.2], [0.2, 0.2, 0.6], [0.5, 0.2, 0.3]]]

    返回结果:
    [[{"label": "Tag", "begin": 0, "end": 1}],
     [{"label": "Tag", "begin": 1, "end": 3}]]

    返回:
    :param batch_sequence_logits: shape: (B, seq_len, label_num) 经过模型预测后得到每一个label的概率分布
    :param mask: shape: (B, seq_len) mask
    :param vocabulary: label的 vocabulary
    :return: decode 之后的 span 字典列表
    """

    if batch_sequence_logits.dim() != 3:
        raise RuntimeError(f"batch_sequence_logits shape 错误, 应该是 (B, seq_len, label_num), "
                           f"而现在是 {batch_sequence_logits.shape}")

    if (mask is not None) and (mask.dim() != 2):
        raise RuntimeError(f"mask shape 错误, 应该是 (B, seq_len), "
                           f"而现在是 {mask.shape}")

    batch = batch_sequence_logits.size(0)

    # mask shape: (B, seq_len)
    if mask is None:
        mask = torch.ones(size=(batch_sequence_logits.shape[0], batch_sequence_logits.shape[1]),
                          dtype=torch.long)

    sequence_length = mask.sum(dim=-1).tolist()

    spans = list()
    for i in range(batch):
        sequence_label, _ = decode_one_sequence_logits_to_label(
            sequence_logits=batch_sequence_logits[i, :sequence_length[i]],
            vocabulary=vocabulary)
        sequence_span = decode_one_sequence_label_to_span(sequence_label)
        spans.append(sequence_span)

    return spans


def decode_label_index_to_span(batch_sequence_label_index: torch.Tensor,
                               mask: torch.ByteTensor,
                               vocabulary: LabelVocabulary) -> List[List[Dict]]:
    """
    将 label index 解码 成span

    batch_sequence_label shape:(B, seq_len)  (B-T: 0, I-T: 1, O: 2)
    [[0, 1, 2],
     [2, 0, 1]]

     对应label序列是:
     [[B, I, O],
      [O, B, I]]

     解码成:

     [[{"label": T, "begin": 0, "end": 2}],
      [{"label": T, "begin": 1, "end": 3}]]

    :param batch_sequence_label_index: shape: (B, seq_len), label index 序列
    :param mask: 对 batch_sequence_label 的 mask
    :param vocabulary: label 词汇表
    :return: 解析好的span列表
    """

    spans = list()

    if mask is None:
        mask = torch.ones(size=(batch_sequence_label_index.shape[0], batch_sequence_label_index.shape[1]),
                          dtype=torch.long)
    mask = mask.bool()
    for sequence_label_index, mask1d in zip(batch_sequence_label_index, mask):
        label_indices = torch.masked_select(sequence_label_index, mask=mask1d).tolist()

        sequence_label = [vocabulary.token(index) for index in label_indices]

        span = decode_one_sequence_label_to_span(sequence_label=sequence_label)
        spans.append(span)

    return spans


def span_intersection(span_list1: List[Dict],
                      span_list2: List[Dict]) -> List[Dict]:
    """
    对 两个 span 列表 求交集
    :param span_list1:
    :param span_list2:
    :return: 交集的结果
    """

    span_set1 = {(span["label"], span["begin"], span["end"]) for span in span_list1}
    span_set2 = {(span["label"], span["begin"], span["end"]) for span in span_list2}

    intersection = [{"label": span[0], "begin": span[1], "end": span[2]} for span in span_set1 & span_set2]

    return intersection


def allowed_transitions(label_vocabulary: LabelVocabulary) -> List[Tuple[int, int]]:
    """
    给定 label 字典，计算 IBO schema 下的 转移矩阵 mask. 因为在 BIO 下面，比如 ["O", "I-Per"]
    或者 ["B-Per", "I-Loc"], 这些转移是不被允许的。该函数，计算mask. 对于允许的转移返回
    pair, (from_label_id, to_label_id)

    这里要特别注意的是: 返回的 allowed pair, 是带有 START 和 STOP的。他们的 index 分别是:
    label_vocabulary.label_size, label_vocabulary.label_size+1

    注意 label_vocabulary.padding_index 也是 label_vocabulary.label_size。虽然是一样的，
    但是因为 START 是用在转移矩阵这个特定场景下，所以不会产生冲突。

    :param label_vocabulary: label 词汇表
    :return: 所有被允许转移的 (from_label_id, to_label_id) pair 对列表。
    """

    num_labels = label_vocabulary.label_size
    start_index = num_labels
    end_index = num_labels + 1

    labels = [(i, label_vocabulary.token(i)) for i in range(label_vocabulary.label_size)]
    labels_with_boundaries = labels + [(start_index, "START"), (end_index, "END")]

    allowed = []
    for from_label_index, from_label in labels_with_boundaries:
        if from_label in ("START", "END"):
            from_tag = from_label
            from_entity = ""
        else:
            from_tag = from_label[0]
            from_entity = from_label[1:]
        for to_label_index, to_label in labels_with_boundaries:
            if to_label in ("START", "END"):
                to_tag = to_label
                to_entity = ""
            else:
                to_tag = to_label[0]
                to_entity = to_label[1:]
            if _is_transition_allowed(from_tag, from_entity, to_tag, to_entity):
                allowed.append((from_label_index, to_label_index))
    return allowed


def _is_transition_allowed(from_tag: str,
                           from_entity: str,
                           to_tag: str,
                           to_entity: str):
    """
    BIO 是否被允许转移。比如: "B-Per" "I-Per" 这是被允许的; 而 "B-Per" "I-Loc" 或者 "O", "I-Per" 这是不被允许的
    :param from_tag: The tag that the transition originates from. For example, if the
        label is ``I-PER``, the ``from_tag`` is ``I``.
    :param from_entity: The entity corresponding to the ``from_tag``. For example, if the
        label is ``I-PER``, the ``from_entity`` is ``PER``.
    :param to_tag: The tag that the transition leads to. For example, if the
        label is ``I-PER``, the ``to_tag`` is ``I``.
    :param to_entity: The entity corresponding to the ``to_tag``. For example, if the
        label is ``I-PER``, the ``to_entity`` is ``PER``.
    :return: True: 该转移是被允许的; False: 该转移是不被允许的。
    """

    if to_tag == "START" or from_tag == "END":
        # Cannot transition into START or from END
        return False

    if from_tag == "START":
        return to_tag in ('O', 'B')

    if to_tag == "END":
        return from_tag in ('O', 'B', 'I')

    return any([
        # Can always transition to O or B-x
        to_tag in ('O', 'B'),
        # Can only transition to I-x from B-x or I-x
        to_tag == 'I' and from_tag in ('B', 'I') and from_entity == to_entity
    ])



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
数据相关定义

Authors: panxu(panxu@baidu.com)
Date:    2020/05/13 15:04:00
"""

from .instance import Instance
from .instance_factory import InstanceFactory
from .vocabulary import Vocabulary, LabelVocabulary
from .model_collate import ModelCollate, ModelInputs
from .pretrained_vocabulary import PretrainedVocabulary
from .pretrained_word_embedding_loader import GeneralPretrainedWordEmbeddingLoader
from .pretrained_word_embedding_loader import PretrainedWordEmbeddingLoader
from .pretrained_word_embedding_loader import GloveLoader
from .pretrained_word_embedding_loader import SGNSLoader



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""

Authors: PanXu
Date:    2020/07/18 16:08:00
"""
import numpy as np
import torch

from easytext.utils.nn.nn_util import masked_softmax


from easytext.tests import ASSERT


def test_masked_softmax():
    """
    测试 masked softmax
    :return:
    """

    vector = torch.FloatTensor(
        [
            [1., 2., 3.],
            [4., 5., 6.]
        ]
    )

    mask = torch.ByteTensor(
        [
            [1, 1, 0],
            [1, 1, 1]
        ]
    )

    result = masked_softmax(vector=vector, mask=mask)

    expect1 = np.exp(np.array([1., 2.]))

    expect1 = expect1 / np.sum(expect1)
    expect1 = np.concatenate([expect1, np.array([0.])], axis=-1).tolist()

    result1 = result[0].tolist()

    ASSERT.assertEqual(len(expect1), len(result1))

    for expect_data, result_data in zip(expect1, result1):
        ASSERT.assertAlmostEqual(expect_data, result_data)

    expect2 = np.exp(np.array([4., 5., 6.]))
    expect2 = expect2 / np.sum(expect2)
    expect2 = expect2.tolist()

    result2 = result[1].tolist()

    ASSERT.assertEqual(len(expect2), len(result2))

    for expect_data, result_data in zip(expect2, result2):
        ASSERT.assertAlmostEqual(expect_data, result_data)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
brief

Authors: PanXu
Date:    2020/11/01 18:00:00
"""
import sys
import importlib
import pkgutil


def import_submodules(package_name: str) -> None:
    """
    导入指定报名的所有包
    :param package_name: 当前包名下所有包
    :return: None
    """
    importlib.invalidate_caches()

    sys.path.append('.')

    module = importlib.import_module(package_name)
    path = getattr(module, '__path__', [])
    path_string = '' if not path else path[0]

    for module_finder, name, _ in pkgutil.walk_packages(path):
        if path_string and module_finder.path != path_string:
            continue
        subpackage = f"{package_name}.{name}"
        import_submodules(subpackage)


#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
对参数的相关梯度进行调整

Authors: PanXu
Date:    2020/08/21 09:34:00
"""

from torch.nn import Module


class GradRescaled:
    """
    对梯度进行调整的接口
    """

    def __call__(self, model: Module):
        """
        对 model 中的参数进行 rescaled 操作
        :param model:
        :return:
        """
        raise NotImplemented()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
trainer 中在训练过程中的 记录

Authors: PanXu
Date:    2020/10/14 10:58:00
"""

from typing import Dict

from easytext.metrics import ModelTargetMetric


class Record:
    """
    训练中的记录数据, 用作 trainer callback 参数
    """

    def __init__(self):
        self.epoch: int = None
        self.epoch_train_num: int = None
        self.epoch_train_loss: float = None
        self.epoch_validation_num: int = None
        self.epoch_validation_loss: float = None

        self.train_metric: Dict = None
        self.train_target_metric: ModelTargetMetric = None
        self.validation_metric: Dict = None
        self.validation_target_metric: ModelTargetMetric = None




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
词汇表

Authors: panxu(panxu@baidu.com)
Date:    2020/04/25 08:33:00
"""
import os
import logging
import json
from typing import List, Dict, Iterable, Union
from collections import Counter


class IVocabulary:
    """
    词汇表的接口类
    """

    # 常量定义
    PADDING = "@@PADDING@@"
    UNK = "@@UNK@@"

    _NEW_LINE = "@@NEWLINE@@"  # "\n" 的替代， 只有在保存字典的时候才会用到

    _VOCABULARY_FILE_NAME = "vocabulary.txt"
    _CONFIG_JSON_FILE_NAME = "config.json"

    def __len__(self):
        raise NotImplementedError()

    @property
    def padding(self) -> str:
        """
        padding属性
        :return: padding属性
        """
        raise NotImplementedError

    @property
    def unk(self):
        """
        unk 属性
        :return: unk 属性
        """
        raise NotImplementedError

    @property
    def padding_index(self) -> Union[None, int]:
        """
        :return: 获取 padding 的 index, 如果 padding 没有设置，那么返回 None; 否则，返回实际的index.
        """
        raise NotImplementedError()

    def index(self, token: str) -> int:
        """
        获取token的index
        :param token: 输入的token
        :return: token 的 index
        """
        raise NotImplementedError()

    def token(self, index: int) -> str:
        """
        获取 index 的 token
        :param index: 指定 index
        :return: 当前 index 的 token
        """
        raise NotImplementedError()

    @property
    def size(self):
        raise NotImplementedError()

    def save_to_file(self, directory: str) -> "Vocabulary":
        """
        保存到指定路径中，会在下面生成两个文件，分别是:
        * vocabulary.txt - 保留全部的字典信息
        * config.json - 保留一些配置信息，比如 padding, unk 等
        :param directory: 存放的路径。
        :return: self
        """

        raise NotImplementedError()

    @classmethod
    def from_file(cls, directory: str) -> "Vocabulary":
        """
        从文件中载入词典，
        保存到指定路径中，包含两个文件，分别是:
        * vocabulary.txt - 保留全部的字典信息
        * config.json - 保留一些配置信息，比如 padding, unk 等
        :param directory: 文件路径
        :return:
        """

        raise NotImplementedError


class Vocabulary(IVocabulary):

    def __init__(self,
                 tokens: Iterable[List[str]],
                 padding: str,
                 unk: str,
                 special_first: bool,
                 other_special_tokens: List = None,
                 min_frequency: int = 1,
                 max_size: int = None):
        """
        初始化
        :param tokens: (B, seq_len)
        :param padding: padding的字符串, 可以用 Vocabulary.PADDING, 如果为 None 或者 "", 表示不进行padding
        :param unk: unknown 的单词，可以用 Vocabulary.UNK, 如果为 None 或者 "", 表示不进行padding
        :param special_first: special 是指: padding, unk.
        True: 表示放在最前面, padding index=0, unk index=1; False: 表示放在最后面。
        这涉及到mask, 对于 token 来说，一般 padding_index = 0;
        而对于 label 来说, 如果需要对label,
        比如 sequence label 进行 padding 的时候, padding_index 和 unk_index 必须大于 label的数量，因为 小于 label 数量的是对应的
        label 分类。
        :param min_frequency: 小于 min_frequency 的会被过滤
        :param max_size: 词表的最大长度, 如果为None, 不限制词表大小
        """
        self._padding = padding
        self._unk = unk
        self.min_frequency = min_frequency
        self.max_size = max_size
        self._token2index: Dict = dict()
        self._index2token: Dict = dict()

        self.other_special_tokens = other_special_tokens or list()

        counter = Counter(t for tt in tokens for t in tt)

        special_tokens = list()

        if self.padding:
            special_tokens.append(self.padding)

        if self.unk:
            special_tokens.append(self.unk)

        special_tokens.extend(self.other_special_tokens)

        # 删除 special tokens
        for st in special_tokens:
            del counter[st]

        # 将counter按照词频从高到低排序
        token_freqency = sorted(counter.items(), key=lambda item: item[1], reverse=True)

        # 构建token index 和 index token 字典

        unique_tokens = list()

        if special_first:
            unique_tokens.extend(special_tokens)

        max_size = None if self.max_size is None else (self.max_size - len(special_tokens))

        for token, freqency in token_freqency:

            if freqency < self.min_frequency:
                continue

            unique_tokens.append(token)

            if max_size and len(unique_tokens) >= max_size:
                break

        if not special_first:
            unique_tokens.extend(special_tokens)

        for index, token in enumerate(unique_tokens):
            self._index2token[index] = token
            self._token2index[token] = index

    def __len__(self):
        return len(self._token2index)

    @property
    def padding(self) -> str:
        return self._padding

    @property
    def unk(self) -> str:
        return self._unk

    @property
    def padding_index(self) -> Union[None, int]:
        """
        :return: 获取 padding 的 index, 如果 padding 没有设置，那么返回 None; 否则，返回实际的index.
        """

        if self.padding is None or self.padding == "":
            return None
        return self.index(self.padding)

    def index(self, token: str) -> int:
        """
        获取token的index
        :param token: 输入的token
        :return: token 的 index
        """
        if token is None or token == "":
            raise RuntimeError(f"token:[{token}] 非法!")

        if not isinstance(token, str):
            raise RuntimeError(f"token 类型: {type(token)} 不是 str")

        if self.unk is None or self.unk == "":
            # 当 unk invalidate 的时候, 说明这里没有 unk, 那么直接返回；
            # 如果 token  有异常，那么，返回错误即可
            index = self._token2index[token]
        else:
            # 如果 unk 存在, 那么, 无效的 token 要返回 unk index.
            index = self._token2index.get(token, self._token2index[self.unk])
        return index

    def token(self, index: int) -> str:
        """
        获取 index 的 token
        :param index: 指定 index
        :return: 当前 index 的 token
        """
        return self._index2token[index]

    @property
    def size(self):
        return len(self._token2index)

    def save_to_file(self, directory: str) -> "Vocabulary":
        """
        保存到指定路径中，会在下面生成两个文件，分别是:
        * vocabulary.txt - 保留全部的字典信息
        * config.json - 保留一些配置信息，比如 padding, unk 等
        :param directory: 存放的路径。
        :return: self
        """

        if not os.path.isdir(directory):
            raise RuntimeError(f"{directory} 不存在!")

        vocabulary_file_path = os.path.join(directory, Vocabulary._VOCABULARY_FILE_NAME)

        if os.path.isfile(vocabulary_file_path):
            logging.warning(f"{vocabulary_file_path} 已经存在，会被覆盖!")

        config_file_path = os.path.join(directory, Vocabulary._CONFIG_JSON_FILE_NAME)

        if os.path.isfile(config_file_path):
            logging.warning(f"{config_file_path} 已经存在，会被覆盖!")

        # 写入 vocabulary
        with open(vocabulary_file_path, mode="w", encoding="utf-8") as f:

            for i in range(len(self._index2token)):
                token = self._index2token[i].replace('\n', Vocabulary._NEW_LINE)
                f.write(f"{token}\n")

        # 写入 config
        with open(config_file_path, mode="w", encoding="utf-8") as f:
            config = {"padding": self.padding,
                      "unk": self.unk,
                      "other_special_tokens": self.other_special_tokens}
            json.dump(config, f, ensure_ascii=False)

        return self

    @classmethod
    def from_file(cls, directory: str) -> "Vocabulary":
        """
        从文件中载入词典，
        保存到指定路径中，包含两个文件，分别是:
        * vocabulary.txt - 保留全部的字典信息
        * config.json - 保留一些配置信息，比如 padding, unk 等
        :param directory: 文件路径
        :return:
        """

        if not os.path.isdir(directory):
            raise RuntimeError(f"{directory} 不存在!")

        vocabulary_file_path = os.path.join(directory, Vocabulary._VOCABULARY_FILE_NAME)

        if not os.path.isfile(vocabulary_file_path):
            raise RuntimeError(f"{vocabulary_file_path} 不存在!")

        config_file_path = os.path.join(directory, Vocabulary._CONFIG_JSON_FILE_NAME)

        if not os.path.isfile(config_file_path):
            raise RuntimeError(f"{config_file_path} 不存在!")

        # 载入config

        with open(config_file_path, mode="r", encoding="utf-8") as f:
            config = json.load(f)

        if config is None:
            raise RuntimeError(f"config 从 {config_file_path} 载入失败!")

        padding = config["padding"]
        unk = config["unk"]
        other_special_tokens = config["other_special_tokens"]

        # 构造对象，实际的数据由后面来填充
        vocabulary = cls([[]],
                         padding=padding,
                         unk=unk,
                         other_special_tokens=other_special_tokens,
                         special_first=True)

        with open(vocabulary_file_path, mode="r", encoding="utf-8") as f:
            for index, line in enumerate(f):
                # 将 \n 去掉，这里不能使用 strip， 因为 token 中可能会存在其他不可见字符
                # 同时将 之前被 _NEW_LINE 替换掉的 "\n" 替换回来
                token = line[:-1].replace(Vocabulary._NEW_LINE, '\n')
                vocabulary._index2token[index] = token
                vocabulary._token2index[token] = index

        return vocabulary


class LabelVocabulary(IVocabulary):
    """
    label 的 vocabulary
    """

    def __init__(self, labels: Iterable[List[str]], padding: str):
        """
        Label 词汇表
        :param labels: label 列表, 二维, 类似于 batch labels
        :param padding: label 需要填充的padding, 填充的 padding index 是不在 [0, label_num) 范围内的，
        也就是说不会影响 label的index. 如果不需要填充，请设置为 `""` 或者 `None`
        """
        self._vocabulary = Vocabulary(tokens=labels,
                                      padding=padding,
                                      unk="",
                                      special_first=False,
                                      min_frequency=1,
                                      max_size=None)

    def __len__(self):
        return len(self._vocabulary)

    @property
    def padding(self) -> str:
        return self._vocabulary.padding

    @property
    def unk(self):
        return self._vocabulary.unk

    @property
    def padding_index(self) -> Union[None, int]:
        return self._vocabulary.padding_index

    def index(self, token: str) -> int:
        return self._vocabulary.index(token)

    def token(self, index: int) -> str:
        return self._vocabulary.token(index)

    @property
    def size(self):
        return self._vocabulary.size

    def save_to_file(self, directory: str) -> "LabelVocabulary":
        self._vocabulary.save_to_file(directory=directory)
        return self

    @classmethod
    def from_file(cls, directory: str) -> "LabelVocabulary":

        label_vocabulary = cls(labels=[[]], padding="")
        label_vocabulary._vocabulary = Vocabulary.from_file(directory=directory)
        return label_vocabulary

    @property
    def label_size(self):
        """
        实际的 label 数量，不包含 padding. 如果需要获得包含 padding 的长度，请使用 size 或者 len(LabelVocabulary)
        :return: label 数量
        """

        if self.padding is None or self.padding == "":
            return len(self)

        else:
            # 将padding 减掉
            return len(self) - 1



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
通过取到最大值的 sequence label 解码

Authors: PanXu
Date:    2020/07/05 11:28:00
"""

from typing import Tuple
import torch

from easytext.utils import bio as BIO
from easytext.data import LabelVocabulary
from .label_index_decoder import LabelIndexDecoder


class SequenceMaxLabelIndexDecoder(LabelIndexDecoder):
    """
    对于 sequence logits, shape: (batch_size, seq_len, num_label), 使用 max 进行 在每一个
    timestep 上进行 decode, 得到 label index.
    """

    def __init__(self, label_vocabulary: LabelVocabulary):
        """
        初始化
        :param label_vocabulary: label 词汇表
        """
        self._label_vocabulary = label_vocabulary

    def __call__(self, logits: torch.Tensor, mask: torch.ByteTensor) -> torch.LongTensor:
        """
        对于 sequence logits, shape: (batch_size, seq_len, num_label), 使用 max 进行 在每一个
        timestep 上进行 decode, 得到 label index.
        :param logits: shape: (batch_size, seq_len, num_label)
        :param mask: shape: (bath_size, seq_len), 存储的是 0 或 1
        :return: 解码后的 label index, shape: (batch_size, seq_len), 注意这是有padding_index 的结果，
        需要使用 mask 来提取实际的 label index.
        """
        if logits.dim() != 3:
            raise RuntimeError(f"logits shape 错误, 应该是 (B, seq_len, num_label), "
                               f"而现在是 {logits.shape}")

        if (mask is not None) and (mask.dim() != 2):
            raise RuntimeError(f"mask shape 错误, 应该是 (B, seq_len), "
                               f"而现在是 {mask.shape}")

        if mask is None:
            mask = torch.ones(size=(logits.size(0), logits.size(1)), dtype=torch.long)

        mask_bool = mask.bool()

        batch_indices = list()
        for sequence_logits, sequence_mask1d in zip(logits, mask_bool):
            # 扩充维度到 2 维
            sequence_mask2d = torch.unsqueeze(sequence_mask1d, dim=-1)

            assert sequence_logits.dim() == sequence_mask2d.dim(), \
                f"sequence_logits dim: {sequence_logits.dim()} 与 sequence_mask.dim: {sequence_mask2d.dim()} 不匹配"

            sequence_logits = torch.masked_select(sequence_logits, mask=sequence_mask2d)
            sequence_logits = sequence_logits.contiguous().view(-1, logits.size(-1))
            sequence_labels, sequence_label_indices = BIO.decode_one_sequence_logits_to_label(
                sequence_logits=sequence_logits,
                vocabulary=self._label_vocabulary)

            sequence_label_indices = torch.tensor(sequence_label_indices, dtype=torch.long, device=logits.device)

            padding = torch.full_like(sequence_mask1d,
                                      fill_value=self._label_vocabulary.padding_index,
                                      dtype=torch.long)

            sequence_label_indices = padding.masked_scatter(sequence_mask1d, sequence_label_indices)
            batch_indices.append(sequence_label_indices)

        batch_indices = torch.stack(batch_indices, dim=0)

        return batch_indices



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
seqeunce 转换成 vector

Authors: PanXu
Date:    2020/07/08 20:09:00
"""

from .cnn_seq2vec import CnnSeq2Vec
from .attention_seq2vec import AttentionSeq2Vec
from .rnn_seq2vec import RnnSeq2Vec



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
对 label 的 f1 metric 计算

Authors: panxu(panxu@baidu.com)
Date:    2020/06/15 19:24:00
"""
from typing import Dict, Union, List
from collections import defaultdict
import torch

from .f1_metric import F1Metric

from easytext.data.vocabulary import Vocabulary


class LabelF1Metric(F1Metric):
    """
    对 label 的 f1 metric 计算. 适用的情况是, label shape: (B,), 每一个 label 是 [0, num_class) 的值。
    f1 的计算是针对 某个 class 的 f1。
    """

    def __init__(self, labels: List[Union[str, int]], label_vocabulary: Vocabulary):
        """
        对 label 的 f1 metric 计算. 适用的情况是, label shape: (B,), 每一个 label 是 [0, num_class) 的值。
        f1 的计算是针对 某个 class 的 f1。

        :param labels: 需要计算 f1 的标签序列. 如果类型是 str, 那么, 需要设置 label_vocabulary，用作
        label 和 index 之间的转换; 如果类型是 int, 那么, 不会使用 label_ vocabulary 进行转换 （可以设置为 None)，
        认为是 label index.
        :param label_vocabulary: label vocabulary 用来将 str label 转换成 label index
        """
        super().__init__(labels=labels)

        if isinstance(labels[0], str) and label_vocabulary is None:
            raise RuntimeError("labels 是 str 类型, label_vocabulary 不能是 None, "
                               "因为需要进行 label 到 index转换")

        self._label_vocabulary = label_vocabulary

        if isinstance(labels[0], str):
            self._label_indices = [label_vocabulary.index(label) for label in self._labels]
        else:
            self._label_indices = labels

    def __call__(self,
                 prediction_labels: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: torch.LongTensor) -> Dict[str, float]:
        """
        计算 metric.

        :param prediction_labels: 预测结果 shape: (B,), 这是模型解码成label的结果，注意是 label，不是 logits
        :param gold_labels: 实际结果 shape: (B,)
        :param mask: mask, shape: (B,)
        :return 每一个label的f1值。这包括 precision, recall, f1。具体结果类似:

        {"precision_[label]": [value],
         "recall_[label]" : [value],
         "f1-measure_[label]": [value],
         "precision-overall": [value],
         "recall-overall": [value],
         "f1-measure-overall": [value]}

         说明: "*-overall" 表示的是所有 命中 在初始化参数的 labels 的综合 metric 值。
         这是有必要的，作为一个综合的值作为统一衡量。
        """

        assert prediction_labels.dim() == 1, "predictions shape 是 (B,)"
        assert gold_labels.dim() == 1, "gold_labels shape 是 (B,)"

        if mask is not None:
            assert mask.dim() == 1, "mask shape 是 (B,)"

        # 转换到 cpu 进行计算
        prediction_labels, gold_labels = prediction_labels.detach().cpu(), gold_labels.detach().cpu()

        if mask is not None:
            bool_mask = mask.detach().cpu().bool()

            prediction_labels = prediction_labels.masked_select(bool_mask)
            gold_labels = gold_labels.masked_select(bool_mask)

        # 当前 batch 下的 true_positives
        true_positives = defaultdict(int)
        false_positives = defaultdict(int)
        false_negatives = defaultdict(int)

        for label, label_index in zip(self._labels, self._label_indices):
            num_prediction = (prediction_labels == label_index).sum().item()
            num_golden = (gold_labels == label_index).sum().item()

            # 计算 true positives
            label_mask = (prediction_labels == label_index)
            label_predictions = prediction_labels.masked_select(label_mask)
            label_gold = gold_labels.masked_select(label_mask)

            true_positives[label] = (label_predictions == label_gold).sum().item()
            false_positives[label] = num_prediction - true_positives[label]
            false_negatives[label] = num_golden - true_positives[label]

        for k, v in true_positives.items():
            self._true_positives[k] += v

        for k, v in false_positives.items():
            self._false_positives[k] += v

        for k, v in false_negatives.items():
            self._false_negatives[k] += v

        return self._metric(true_positives=true_positives,
                            false_positives=false_positives,
                            false_negatives=false_negatives)




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
测试 Bio

Authors: panxu(panxu@baidu.com)
Date:    2020/05/15 10:50:00
"""
import torch
from easytext.data import LabelVocabulary
from easytext.utils import bio as BIO

from easytext.tests import ASSERT


def test_fill():
    """
    测试 bio
    :return:
    """

    pairs = [(1, 2), (2, 4)]

    for begin, end in pairs:
        sl = ["O"] * 10
        tag = "Test"
        BIO.fill(sequence_label=sl, begin_index=begin, end_index=end, tag=tag)

        for i in range(begin, end):

            if i == begin:
                ASSERT.assertEqual(sl[i], f"B-{tag}")
            else:
                ASSERT.assertEqual(sl[i], f"I-{tag}")


def test_decode_one_sequence_logits_to_label():
    """
    测试 decode sequence label
    :return:
    """

    sequence_logits_list = list()
    expect_list = list()

    sequence_logits = torch.tensor([[0.2, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                   dtype=torch.float)  # O B I 正常
    expect = ["O", "B-T", "I-T"]
    sequence_logits_list.append(sequence_logits)
    expect_list.append(expect)

    sequence_logits = torch.tensor([[0.9, 0.3, 0.4], [0.2, 0.8, 0.3], [0.2, 0.3, 0.1]],
                                   dtype=torch.float)
    expect = ["B-T", "I-T", "I-T"]

    sequence_logits_list.append(sequence_logits)
    expect_list.append(expect)

    sequence_logits = torch.tensor([[0.9, 0.3, 0.4], [0.2, 0.8, 0.3], [0.2, 0.3, 0.9]],
                                   dtype=torch.float)
    expect = ["B-T", "I-T", "O"]
    sequence_logits_list.append(sequence_logits)
    expect_list.append(expect)

    vocabulary = LabelVocabulary([["B-T", "B-T", "B-T", "I-T", "I-T", "O"]],
                                 padding=LabelVocabulary.PADDING)

    b_index = vocabulary.index("B-T")
    ASSERT.assertEqual(0, b_index)
    i_index = vocabulary.index("I-T")
    ASSERT.assertEqual(1, i_index)
    o_index = vocabulary.index("O")
    ASSERT.assertEqual(2, o_index)

    for sequence_logits, expect in zip(sequence_logits_list, expect_list):
        sequence_label, sequence_label_indices = BIO.decode_one_sequence_logits_to_label(
            sequence_logits=sequence_logits,
            vocabulary=vocabulary)

        ASSERT.assertListEqual(sequence_label, expect)
        expect_indices = [vocabulary.index(label) for label in expect]
        ASSERT.assertListEqual(sequence_label_indices, expect_indices)


def test_decode_one_sequence_logits_to_label_abnormal():
    """
    测试异常case
    :return:
    """

    # [0.2, 0.5, 0.4] argmax 解码是 I 这是异常的 case, 整个序列是: I B O
    # 而 decode_sequence_lable_bio 会将 概率值 是 0.4 的 也就是 O 作为标签输出 来修订这个个错误
    sequence_logits = torch.tensor([[0.2, 0.5, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                   dtype=torch.float)

    vocabulary = LabelVocabulary([["B-T", "B-T", "B-T", "I-T", "I-T", "O"]],
                                 padding=LabelVocabulary.PADDING)

    b_index = vocabulary.index("B-T")
    ASSERT.assertEqual(0, b_index)
    i_index = vocabulary.index("I-T")
    ASSERT.assertEqual(1, i_index)
    o_index = vocabulary.index("O")
    ASSERT.assertEqual(2, o_index)

    sequence_label, sequence_label_indices = BIO.decode_one_sequence_logits_to_label(sequence_logits=sequence_logits,
                                                             vocabulary=vocabulary)

    expect = ["O", "B-T", "I-T"]
    expect_indices = [vocabulary.index(label) for label in expect]
    ASSERT.assertListEqual(expect, sequence_label)
    ASSERT.assertListEqual(expect_indices, sequence_label_indices)

    # argmax 解码是 I I I 经过修订后是: O O B
    sequence_logits = torch.tensor([[0.2, 0.5, 0.4], [0.2, 0.9, 0.3], [0.2, 0.3, 0.1]],
                                   dtype=torch.float)

    sequence_label, sequence_label_indices = BIO.decode_one_sequence_logits_to_label(sequence_logits=sequence_logits,
                                                             vocabulary=vocabulary)
    expect = ["O", "O", "B-T"]
    expect_indices = [vocabulary.index(label) for label in expect]
    ASSERT.assertListEqual(expect, sequence_label)
    ASSERT.assertListEqual(expect_indices, sequence_label_indices)


def test_decode_one_sequence_label_to_span():
    """
    测试对 sequence label 解码成 span 字典
    :return:
    """

    sequence_label_list = list()
    expect_list = list()

    sequence_label = ["B-T", "I-T", "O-T"]
    expect = [{"label": "T", "begin": 0, "end": 2}]

    sequence_label_list.append(sequence_label)
    expect_list.append(expect)

    sequence_label = ["B-T", "I-T", "I-T"]
    expect = [{"label": "T", "begin": 0, "end": 3}]

    sequence_label_list.append(sequence_label)
    expect_list.append(expect)

    sequence_label = ["B-T", "I-T", "I-T", "B-T"]
    expect = [{"label": "T", "begin": 0, "end": 3},
              {"label": "T", "begin": 3, "end": 4}]

    sequence_label_list.append(sequence_label)
    expect_list.append(expect)

    for expect, sequence_label in zip(expect_list, sequence_label_list):
        span = BIO.decode_one_sequence_label_to_span(sequence_label)

        ASSERT.assertListEqual(expect, span)


def test_decode():
    """
    测试 模型输出的 batch logits 解码
    :return:
    """

    # [[O, B, I], [B, B, I], [B, I, I], [B, I, O]]
    batch_sequence_logits = torch.tensor([[[0.2, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.7, 0.2, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.1, 0.7, 0.3], [0.2, 0.3, 0.1]],
                                          [[0.8, 0.3, 0.4], [0.1, 0.7, 0.3], [0.2, 0.3, 0.5]]],
                                         dtype=torch.float)

    expect = [[{"label": "T", "begin": 1, "end": 3}],
              [{"label": "T", "begin": 0, "end": 1}, {"label": "T", "begin": 1, "end": 3}],
              [{"label": "T", "begin": 0, "end": 3}],
              [{"label": "T", "begin": 0, "end": 2}]]

    vocabulary = LabelVocabulary([["B-T", "B-T", "B-T", "I-T", "I-T", "O"]],
                                 padding=LabelVocabulary.PADDING)

    b_index = vocabulary.index("B-T")
    ASSERT.assertEqual(0, b_index)
    i_index = vocabulary.index("I-T")
    ASSERT.assertEqual(1, i_index)
    o_index = vocabulary.index("O")
    ASSERT.assertEqual(2, o_index)

    spans = BIO.decode(batch_sequence_logits=batch_sequence_logits,
                       mask=None,
                       vocabulary=vocabulary)

    ASSERT.assertListEqual(expect, spans)


def test_decode_decode_label_index_to_span():
    """
    测试解码 golden label index
    :return:
    """

    vocabulary = LabelVocabulary([["B-T", "B-T", "B-T", "I-T", "I-T", "O"]],
                                 padding=LabelVocabulary.PADDING)

    b_index = vocabulary.index("B-T")
    ASSERT.assertEqual(0, b_index)
    i_index = vocabulary.index("I-T")
    ASSERT.assertEqual(1, i_index)
    o_index = vocabulary.index("O")
    ASSERT.assertEqual(2, o_index)

    golden_labels = torch.tensor([[0, 1, 2, 0],
                                  [2, 0, 1, 1]])

    expect = [[{"label": "T", "begin": 0, "end": 2}, {"label": "T", "begin": 3, "end": 4}],
              [{"label": "T", "begin": 1, "end": 4}]]

    spans = BIO.decode_label_index_to_span(batch_sequence_label_index=golden_labels,
                                           mask=None,
                                           vocabulary=vocabulary)

    ASSERT.assertListEqual(expect, spans)


def test_span_intersection():
    span_list1 = [{"label": "T", "begin": 0, "end": 2}, {"label": "T", "begin": 3, "end": 4}]
    span_list2 = [{"label": "T", "begin": 0, "end": 2}]

    intersetction = BIO.span_intersection(span_list1=span_list1,
                                          span_list2=span_list2)

    expect = [{"label": "T", "begin": 0, "end": 2}]
    ASSERT.assertListEqual(expect, intersetction)

    span_list1 = [{"label": "T", "begin": 0, "end": 2}, {"label": "T", "begin": 3, "end": 4}]
    span_list2 = [{"label": "T", "begin": 9, "end": 10}]

    intersetction = BIO.span_intersection(span_list1=span_list1,
                                          span_list2=span_list2)

    expect = []
    ASSERT.assertListEqual(expect, intersetction)

    span_list1 = [{"label": "T", "begin": 0, "end": 2}, {"label": "T", "begin": 3, "end": 4}]
    span_list2 = []

    intersetction = BIO.span_intersection(span_list1=span_list1,
                                          span_list2=span_list2)

    expect = []
    ASSERT.assertListEqual(expect, intersetction)


def test_allowed_transitions():
    """
    测试允许转移mask pair
    :return:
    """

    label_vocabulary = LabelVocabulary(labels=[["B-L1", "I-L1", "B-L2", "I-L2", "O"]],
                                       padding=LabelVocabulary.PADDING)

    allowed_pairs = BIO.allowed_transitions(label_vocabulary=label_vocabulary)

    for from_idx, to_idx in allowed_pairs:

        if from_idx == label_vocabulary.label_size:
            from_label = "START"
        else:
            from_label = label_vocabulary.token(from_idx)

        if to_idx == label_vocabulary.label_size + 1:
            to_label = "STOP"
        else:
            to_label = label_vocabulary.token(to_idx)
        print(f"(\"{from_label}\", \"{to_label}\"),")

    expect_trainsition_labels = [
        ("B-L1", "B-L1"), ("B-L1", "I-L1"), ("B-L1", "B-L2"), ("B-L1", "O"), ("B-L1", "STOP"),
        ("I-L1", "B-L1"), ("I-L1", "I-L1"), ("I-L1", "B-L2"), ("I-L1", "O"), ("I-L1", "STOP"),
        ("B-L2", "B-L1"), ("B-L2", "B-L2"), ("B-L2", "I-L2"), ("B-L2", "O"), ("B-L2", "STOP"),
        ("I-L2", "B-L1"), ("I-L2", "B-L2"), ("I-L2", "I-L2"), ("I-L2", "O"), ("I-L2", "STOP"),
        ("O", "B-L1"), ("O", "B-L2"), ("O", "O"), ("O", "STOP"),
        ("START", "B-L1"), ("START", "B-L2"), ("START", "O")]


    expect = list()

    for from_label, to_label in expect_trainsition_labels:
        if from_label == "START":
            from_idx = label_vocabulary.label_size
        else:
            from_idx = label_vocabulary.index(from_label)

        if to_label == "STOP":
            to_idx = label_vocabulary.label_size + 1
        else:
            to_idx = label_vocabulary.index(to_label)

        expect.append((from_idx, to_idx))

    ASSERT.assertSetEqual(set(expect), set(allowed_pairs))



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 label index decoder

Authors: PanXu
Date:    2020/07/05 15:10:00
"""
import pytest
import torch

from easytext.tests import ASSERT

from easytext.data import LabelVocabulary
from easytext.modules import ConditionalRandomField
from easytext.label_decoder import CRFLabelIndexDecoder


class CRFData:
    """
    测试用的 crf 数据
    """

    def __init__(self):
        bio_labels = [["O", "I-X", "B-X", "I-Y", "B-Y"]]

        self.label_vocabulary = LabelVocabulary(labels=bio_labels,
                                                padding=LabelVocabulary.PADDING)

        self.logits = torch.tensor([
            [[0, 0, .5, .5, .2], [0, 0, .3, .3, .1], [0, 0, .9, 10, 1]],
            [[0, 0, .2, .5, .2], [0, 0, 3, .3, .1], [0, 0, .9, 1, 1]],
        ], dtype=torch.float)

        self.tags = torch.tensor([
            [2, 3, 4],
            [3, 2, 2]
        ], dtype=torch.long)

        self.transitions = torch.tensor([
            [0.1, 0.2, 0.3, 0.4, 0.5],
            [0.8, 0.3, 0.1, 0.7, 0.9],
            [-0.3, 2.1, -5.6, 3.4, 4.0],
            [0.2, 0.4, 0.6, -0.3, -0.4],
            [1.0, 1.0, 1.0, 1.0, 1.0]
        ], dtype=torch.float)

        self.transitions_from_start = torch.tensor([0.1, 0.2, 0.3, 0.4, 0.6], dtype=torch.float)
        self.transitions_to_end = torch.tensor([-0.1, -0.2, 0.3, -0.4, -0.4], dtype=torch.float)

        # Use the CRF Module with fixed transitions to compute the log_likelihood
        self.crf = ConditionalRandomField(5)
        self.crf.transitions = torch.nn.Parameter(self.transitions)
        self.crf.start_transitions = torch.nn.Parameter(self.transitions_from_start)
        self.crf.end_transitions = torch.nn.Parameter(self.transitions_to_end)

        # constraint crf
        constraints = {(0, 0), (0, 1),
                       (1, 1), (1, 2),
                       (2, 2), (2, 3),
                       (3, 3), (3, 4),
                       (4, 4), (4, 0)}

        # Add the transitions to the end tag
        # and from the start tag.
        for i in range(5):
            constraints.add((5, i))
            constraints.add((i, 6))

        constraint_crf = ConditionalRandomField(num_tags=5, constraints=constraints)
        constraint_crf.transitions = torch.nn.Parameter(self.transitions)
        constraint_crf.start_transitions = torch.nn.Parameter(self.transitions_from_start)
        constraint_crf.end_transitions = torch.nn.Parameter(self.transitions_to_end)
        self.constraint_crf = constraint_crf


@pytest.fixture(scope="class")
def crf_data():
    """
    产生测试用的 crf data
    :return:
    """
    return CRFData()


def test_crf_label_index_decoder(crf_data):
    """
    测试 crf label index decoder
    :param crf_data: crf data
    :return:
    """
    mask = torch.tensor([
        [1, 1, 1],
        [1, 1, 0]
    ], dtype=torch.long)

    crf_label_index_decoder = CRFLabelIndexDecoder(crf=crf_data.crf,
                                                   label_vocabulary=crf_data.label_vocabulary)

    label_indices = crf_label_index_decoder(logits=crf_data.logits,
                                            mask=mask)
    padding_index = crf_data.label_vocabulary.padding_index
    expect = [[2, 4, 3], [4, 2, padding_index]]

    ASSERT.assertListEqual(expect, label_indices.tolist())


def test_crf_label_index_decoder_with_constraint(crf_data):
    mask = torch.tensor([
        [1, 1, 1],
        [1, 1, 0]
    ], dtype=torch.uint8)

    crf_label_index_decoder = CRFLabelIndexDecoder(crf=crf_data.constraint_crf,
                                                   label_vocabulary=crf_data.label_vocabulary)

    label_indices = crf_label_index_decoder(logits=crf_data.logits,
                                            mask=mask)
    padding_index = crf_data.label_vocabulary.padding_index
    expect = [[2, 3, 3], [2, 3, padding_index]]

    ASSERT.assertListEqual(expect, label_indices.tolist())



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
循环神经网络 sequence encoding 成 vector

Authors: PanXu
Date:    2020/10/20 17:19:00
"""

from typing import Dict

from torch.nn import Module
from torch import Tensor, BoolTensor

from easytext.modules import DynamicRnn, DynamicRnnOutput


class RnnSeq2Vec(Module):
    """
    rnn seq2vec, batch first
    """

    def __init__(self, dynamic_rnn: DynamicRnn):
        super().__init__()
        self.rnn = dynamic_rnn

    def forward(self, sequence: Tensor, mask: BoolTensor) -> Dict[str, Tensor]:
        rnn_output: DynamicRnnOutput = self.rnn(sequence=sequence,
                                                mask=mask)

        return rnn_output.last_layer_h_n



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2021 PanXu, Inc. All Rights Reserved
#
"""
图注意力模型

Authors: PanXu
Date:    2021/02/14 10:00:00
"""

import torch
from torch import nn
from torch.nn import Dropout, Parameter, Linear, LeakyReLU, ModuleList
from torch.nn import ELU, LogSoftmax
import torch.nn.functional as F


class GraphAttentionLayer(nn.Module):
    """
    图 Attention Layer
    """

    def __init__(self,
                 in_features: int,
                 out_features: int,
                 dropout: float,
                 alpha: float):
        """
        初始化
        :param in_features: 输入的特征数量
        :param out_features: 输出的特征数
        :param dropout: dropout
        :param alpha: leakyrelu 的 alpha 参数
        """
        super().__init__()

        self.dropout = Dropout(dropout)
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha

        self.feed_forward = Linear(in_features, out_features, bias=False)

        self.a1 = Parameter(torch.Tensor(out_features, 1))
        self.a2 = Parameter(torch.Tensor(out_features, 1))

        self.leaky_relu = LeakyReLU(self.alpha)

        self.reset_parameters()

    def reset_parameters(self):
        """
        初始化参数
        :return:
        """
        nn.init.xavier_uniform_(self.feed_forward.weight, gain=1.414)
        nn.init.xavier_uniform_(self.a1, gain=1.414)
        nn.init.xavier_uniform_(self.a2, gain=1.414)

    def forward(self, input: torch.Tensor, adj: torch.Tensor) -> torch.Tensor:
        """
        模型运算
        :param input: (B, num_nodes, in_features)
        :param adj: 邻接矩阵 (B, num_nodes, num_nodes)
        :return: (B, node_len, out_features)
        """

        # 线性变换, 转换成隐层, (B, node_len, in_features) -> (B, node_len, out_features)
        h = self.feed_forward(input)

        # [batch_size, N, out_features]
        batch_size, num_nodes, _ = h.size()

        # 这里是将 a \cdot [Wf_i || W_fj] 这样的运算拆开来做了, 这就与 GAT 实际算法描述是一致的了
        #  a \cdot [Wf_i || W_fj] = a_1 * Wf_i + a_2 * Wf_j
        middle_result1 = torch.matmul(h, self.a1).expand(-1, -1, num_nodes)
        middle_result2 = torch.matmul(h, self.a2).expand(-1, -1, num_nodes).transpose(1, 2)

        # 计算 leaky_relu
        e = self.leaky_relu(middle_result1 + middle_result2)

        # 计算 attetion, 注意 为了使用 softmax, 所以这里将 邻接矩阵为 0 的，用极大的负值填充
        # 这样就会在计算 softmax 中为 0， 因为 exp(float("-inf"))=0
        attention1 = e.masked_fill(adj == 0, -1e9)

        # 计算 attention softmax
        attention = F.softmax(attention1, dim=2)
        attention = self.dropout(attention)

        h_prime = torch.matmul(attention, h)

        return h_prime

    def __repr__(self):
        return self.__class__.__name__ + ' (' + str(self.in_features) + ' -> ' + str(self.out_features) + ')'


class GAT(nn.Module):
    """
    图注意力模型，当前模型，最多支持两层
    """

    def __init__(self,
                 in_features: int,
                 out_features: int,
                 dropout: float,
                 alpha: float,
                 num_heads: int,
                 hidden_size: int = None):
        """
        初始化
        :param in_features: 输入的 node 维度
        :param out_features: 输出的 node 维度
        :param dropout: dropout
        :param alpha: 在 GraphAttentionLayer 中 LeakyRelu 用到的 alpha
        :param num_heads: 头的数量
        :param hidden_size: 隐层 size，如果是 None 表示没有隐层; 否则，只有一个隐层
        """

        super().__init__()
        self.dropout = Dropout(dropout)
        self.hidden_size = hidden_size
        self.activation = ELU()
        self.log_softmax = LogSoftmax(dim=2)
        if hidden_size is None:
            self.layers = ModuleList([GraphAttentionLayer(in_features=in_features,
                                                          out_features=out_features,
                                                          dropout=dropout,
                                                          alpha=alpha)
                                      for _ in range(num_heads)])
        else:
            self.layers = ModuleList([GraphAttentionLayer(in_features=in_features,
                                                          out_features=hidden_size,
                                                          dropout=dropout,
                                                          alpha=alpha)
                                      for _ in range(num_heads)])
            self.final_layer = GraphAttentionLayer(in_features=hidden_size * num_heads,
                                                   out_features=out_features,
                                                   dropout=dropout,
                                                   alpha=alpha)

    def forward(self, nodes: torch.Tensor, adj: torch.Tensor) -> torch.Tensor:
        """
        GAT 运算
        :param nodes: 图的节点，shape: (B, num_nodes, in_features)
        :param adj: 图的邻接矩阵，不包含 self loop
        :return: 计算后的结果, shape: (B, num_nodes, output_features)
        """

        assert nodes.dim() == 3, f"nodes 的维度: {nodes.dim()}, 与 (B, num_nodes, in_features) 不匹配"

        nodes = self.dropout(nodes)

        if self.hidden_size is None:
            nodes = torch.stack([self.activation(layer(nodes, adj)) for layer in self.layers], dim=2)
            nodes = nodes.sum(2)
            nodes = self.dropout(nodes)
        else:
            nodes = torch.cat([self.activation(layer(nodes, adj)) for layer in self.layers], dim=2)
            nodes = self.dropout(nodes)
            nodes = self.activation(self.final_layer(nodes, adj))
        return self.log_softmax(nodes)




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2021 PanXu, Inc. All Rights Reserved
#
"""
基于 <<Chinese NER Using Lattice LSTM>> 论文
论文地址: https://www.aclweb.org/anthology/P18-1144/

实现 lattice lstm 模型
相关说明文档参考:

docs/ner/Chinese NER Using Lattice LSTM.md

Authors: PanXu
Date:    2021/01/20 19:48:00
"""
import logging
from typing import Tuple, List, Dict

import numpy as np

import torch
from torch import nn
from torch.nn import init
from torch.nn import Module, Parameter


class WordLSTMCell(Module):
    """
    word cell

    相关说明文档: docs/ner/Chinese NER Using Lattice LSTM.md

    Part1 中，计算的结果，运算得到 c^w_{b,e}
    """

    def __init__(self, input_size, hidden_size, bias=True):
        """
        初始化
        :param input_size: w_{b,e} 也就是词向量的维度
        :param hidden_size: 输出的隐层维度, 注意实际会 hidden_size*3
        :param bias: 是否有 bias
        """

        super(WordLSTMCell, self).__init__()

        # input size
        self.input_size = input_size

        # 输出的 hidden size
        self.hidden_size = hidden_size

        # 是否使用 bias
        self.use_bias = bias

        # W*[x^w_{b,e}; h^c_b] + b = weight_ih*(x^w_{b,e}) + weight_hh*(h^c_b) + b 计算过程
        # weight_ih*(x^w_{b,e}) 计算该部分的参数
        # 注意: 3 * hidden_size 是因为一次性将 i, f, o 三个值计算出来，再通过 split 分开得到 3 个值，所以需要乘以 3
        self.weight_ih = nn.Parameter(
            torch.FloatTensor(input_size, 3 * hidden_size))

        # weight_hh*(h^c_b) 计算该部分的参数
        self.weight_hh = nn.Parameter(
            torch.FloatTensor(hidden_size, 3 * hidden_size))

        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(3 * hidden_size))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        """
        reset 参数
        :return:
        """
        init.orthogonal(self.weight_ih)

        weight_hh_data = torch.eye(self.hidden_size)
        weight_hh_data = weight_hh_data.repeat(1, 3)

        with torch.no_grad():
            self.weight_hh.copy_(weight_hh_data)

        if self.bias is not None:
            init.constant(self.bias, val=0)

    def forward(self, input_, hx) -> torch.Tensor:
        """
        Args:
            input_: 是词向量，也就是 x^w_{b,e} 的词向量， size: (B, input_size)
            hx: (h_0, c_0), 是 h^c_b, 也就是在 b 处的隐层输出向量，size: (B, hidden_size).
        Returns:
            c_1: 是 part1 部分计算的结果，是 c^w_{b,e} 该值, size: (B, hidden_size)
        """

        # h_0: h^c_b 也就是 [b,e] 的 b 所在的 h
        h_0, c_0 = hx
        batch_size = h_0.size(0)

        if self.bias is not None:
            bias_batch = (self.bias.unsqueeze(0).expand(batch_size, *self.bias.size()))

        # weight_hh * h^c_b + b

        if self.bias is not None:
            wh_b = torch.addmm(bias_batch, h_0, self.weight_hh)
        else:
            wh_b = torch.mm(h_0, self.weight_hh)

        # weight_ih * x^w_{b,e}
        wi = torch.mm(input_, self.weight_ih)

        # 计算 f, i, g
        f, i, g = torch.split(wh_b + wi, split_size_or_sections=self.hidden_size, dim=1)

        # 最终计算出 c^w_{b,e}
        c_1 = torch.sigmoid(f)*c_0 + torch.sigmoid(i)*torch.tanh(g)
        return c_1

    def __repr__(self):
        s = '{name}({input_size}, {hidden_size})'
        return s.format(name=self.__class__.__name__, **self.__dict__)


class MultiInputLSTMCell(nn.Module):

    """
    结合 WordLSTMCell 和 LSTM Cell 的计算.

    Part2 和 Part3 计算, 计算得到多个 i, 并将 word (WordLSTMCell 计算的结果) 与 char 的 c_t 合并在一起。
    """

    def __init__(self, input_size, hidden_size, bias=True):
        """
        初始化
        :param input_size: input_size, 是指输入的 字 的 embedding 维度，也就是 x^c_j
        :param hidden_size: 输出的隐层维度
        :param bias: True: 使用 bias; False: 不使用 bias
        """

        super().__init__()

        # 输入的 embedding 维度
        self.input_size = input_size

        # 输出的 隐层 embedding 维度
        self.hidden_size = hidden_size

        # 与 x^c_j 进行相乘的参数
        self.weight_ih = nn.Parameter(
            torch.FloatTensor(input_size, 3 * hidden_size))

        # 与 h^c_{j-1} 进行相乘的参数
        self.weight_hh = nn.Parameter(
            torch.FloatTensor(hidden_size, 3 * hidden_size))

        # 用来计算 x^e_{b,e}(WordLSTMCell 计算出的结果) 与 当前 x^c_j 的系数的参数, 这是与 x^c_j 乘积的部分
        self.alpha_weight_ih = nn.Parameter(
            torch.FloatTensor(input_size, hidden_size))

        # 用来计算 x^e_{b,e}(WordLSTMCell 计算出的结果) 与 当前 x^c_j 的系数的参数, 这是与 x^e_{b,e} 乘积的部分
        self.alpha_weight_hh = nn.Parameter(
            torch.FloatTensor(hidden_size, hidden_size))

        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(3 * hidden_size))
            self.alpha_bias = nn.Parameter(torch.FloatTensor(hidden_size))
        else:
            self.register_parameter('bias', None)
            self.register_parameter('alpha_bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        """
        reset 参数
        :return:
        """
        init.orthogonal(self.weight_ih)
        init.orthogonal(self.alpha_weight_ih)

        weight_hh_data = torch.eye(self.hidden_size)
        weight_hh_data = weight_hh_data.repeat(1, 3)

        with torch.no_grad():
            self.weight_hh.copy_(weight_hh_data)

        alpha_weight_hh_data = torch.eye(self.hidden_size)
        alpha_weight_hh_data = alpha_weight_hh_data.repeat(1, 1)

        with torch.no_grad():
            self.alpha_weight_hh.copy_(alpha_weight_hh_data)

        # The bias is just set to zero vectors.
        if self.bias is not None:
            init.constant(self.bias, val=0)
            init.constant(self.alpha_bias, val=0)

    def forward(self, input_, c_input, hx) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        注意: 当前模型仅仅支持 batch_size = 1
        :param input_: 字 embedding 输入向量
        :param c_input: 由 WordLSTMCell 计算得出的 所有 词的向量。
        :param hx: 在 j-1 步的输出, h_{j-1}, c_{j-1}
        :return: h_{j}, c_{j} 当前 cell 输出的隐层 和 cell 输出
        """

        # h^c_{j-1}, c^c_{j-1}, 前一个输出的 h, c, 该函数运算 lstm 一个 cell, 并得到 h^c_j, c^c_j
        h_0, c_0 = hx

        batch_size = h_0.size(0)

        # 注意只能处理 batch_size 为 1 的情况
        assert(batch_size == 1)

        if self.bias is not None:
            bias_batch = (self.bias.unsqueeze(0).expand(batch_size, *self.bias.size()))

        # $W*[x^c_j;h^c_{j-1}] + b = (weight_hh * h^c_{j-1}) + (weight_ih * x^c_j) + b$
        # $(weight_hh * h^c_{j-1}) + b$
        if self.bias is not None:
            wh_b = torch.addmm(bias_batch, h_0, self.weight_hh)
        else:
            wh_b = torch.mm(h_0, self.weight_hh)

        # weight_ih * x^c_j
        wi = torch.mm(input_, self.weight_ih)
        # 计算 i, o, g, g 就是 $\tilde{c_j}$
        i, o, g = torch.split(wh_b + wi, split_size_or_sections=self.hidden_size, dim=1)

        # 计算 i, o, g
        i = torch.sigmoid(i)
        g = torch.tanh(g)
        o = torch.sigmoid(o)

        # c_sum, 是指当前j为结尾，一共命中了多少词
        c_num = len(c_input)

        if c_num == 0:
            # 没有命中词，则使用常规的 lstm 方法进行处理
            f = 1 - i
            c_1 = f*c_0 + i*g
            h_1 = o * torch.tanh(c_1)
        else:
            # 命中了多个词
            # 将所有命中词的向量，组合成一个向量, 按照0维合并
            c_input_var = torch.cat(c_input, 0)

            # 缩减维度，去掉 batch_size
            c_input_var = c_input_var.squeeze(1)

            # 计算 part3 中的 $i^c_{b,e}$
            # i^c_{b,e} = W*[x^c_j;c^w_{b,e}] + b = alpha_weight_ih*x^c_j + alpha_weight_hh * c^w_{b,e}
            # 其中 c^w_{b,e} 是在 WordLSTMCell 中计算的结果

            if self.alpha_bias is not None:
                alpha_wi = torch.addmm(self.alpha_bias, input_, self.alpha_weight_ih).expand(c_num, self.hidden_size)
            else:
                alpha_wi = torch.mm(input_, self.alpha_weight_ih).expand(c_num, self.hidden_size)

            alpha_wh = torch.mm(c_input_var, self.alpha_weight_hh)

            # alpha 就是 i^c_{b,e}
            alpha = torch.sigmoid(alpha_wi + alpha_wh)

            # 将所有的 i^c_j 与 所有的 i^c_{b,e} 组合在一起, 进行 softmax 计算
            alpha = torch.exp(torch.cat([i, alpha],0))
            alpha_sum = alpha.sum(0)
            alpha = torch.div(alpha, alpha_sum)

            # 最后一步，将 g=$\tilde{c_j}$, 以及 所有 c^w_{b,e} 放在一起，分别乘以权重进行计算
            merge_i_c = torch.cat([g, c_input_var],0)

            # 分别乘以权重，得到 c_1
            c_1 = merge_i_c * alpha
            c_1 = c_1.sum(0).unsqueeze(0)

            # 与常规 lstm 一样计算 h_1
            h_1 = o * torch.tanh(c_1)
        return h_1, c_1

    def __repr__(self):
        s = '{name}({input_size}, {hidden_size})'
        return s.format(name=self.__class__.__name__, **self.__dict__)


class LatticeLSTM(nn.Module):
    """
    基于 MultiInputLSTMCell 的 LSTM 模型
    """

    def __init__(self,
                 input_dim: int,
                 hidden_dim: int,
                 gaz_word_embedding_dim: int,
                 gaz_word_embedding: torch.Tensor,
                 gaz_word_embedding_dropout: float,
                 left2right: bool):
        """
        Lattice 初始出啊
        :param input_dim: 输入的维度
        :param hidden_dim: 隐层输出的维度
        :param gaz_word_embedding_dim: gaz 词向量的维度
        :param gaz_word_embedding: gaz 的词向量
        :param gaz_word_embedding_dropout: gaz 词向量的 dropout
        :param left2right: 从左向右 还是 从右向左，就是语言模型的两个方向
        """

        super().__init__()

        self.hidden_dim = hidden_dim
        self.word_emb = gaz_word_embedding

        self.word_dropout = nn.Dropout(gaz_word_embedding_dropout)

        self.rnn = MultiInputLSTMCell(input_dim, hidden_dim)
        self.word_rnn = WordLSTMCell(gaz_word_embedding_dim, hidden_dim)

        self.left2right = left2right

        self.reset_parameters()

    def reset_parameters(self):
        pass

    def forward(self,
                input: torch.Tensor,
                skip_input: List[List[List[List]]],
                hidden: Tuple[torch.Tensor, torch.Tensor] = None) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        执行模型
        :param input: 字序列向量, shape: (B, seq_len, embedding_size),
                      但是 batch_size = 1， 必须是 1，也就是不支持其他 batch_size
        :param skip_input: skip_input: 是4维 list, 因为 B=1, 所以内部是一个 3维list,
                           该 3维 list 总长度是 seq_len(与字符序列是一样长度的).
                           每一个元素，是对应到词汇表中词的 id 和 对应的长度。例如:
                           [[], [[25,13],[2,4]], [], [[33], [2]], []], 表示在字序列中，第 2个 字，所对应的词 id 是 25 和13 , 对应的长度是 2 和 4。
                           例如: "到 长 江 大 桥", 该序列长度是 5， 所以 skip_input 也是 5, 其中 "长" index=1,
                           对应 "长江" 和 "长江大桥", 其中 "长江" 在词汇表中的 id 是25, 长度是 2;
                           "长江大桥" 对应词汇表中 id 是 13， 长度是 4;
                            同样 "大桥", 对应 词汇表 id 33, 长度是 2.
        :param hidden: 预定义的 (h,c) 输入
        :return: (h1, c1), ..., (hn, cn), 返回的是 sequence 隐层序列, hj 或 cj 的 shape: (B, seq_len, hidden_dim), 其中 B=1
        """

        input = input.transpose(1, 0)

        seq_len = input.size(0)
        batch_size = input.size(1)

        # 只能处理 batch_size 为 1
        assert (batch_size == 1)

        # 因为 batch_size == 1, 所以仅仅取第一个
        skip_input = skip_input[0]

        if not self.left2right:  # 如果是 right2left 需要将 构成的词汇表也进行逆向
            skip_input = convert_forward_gaz_to_backward(skip_input)

        hidden_out = []  # h
        memory_out = []  # c

        if hidden:  # 如果 hidden_{t-1} 存在，则使用; 否则，使用 0
            (hx, cx) = hidden
        else:
            hx = torch.zeros(batch_size, self.hidden_dim, device=input.device)
            cx = torch.zeros(batch_size, self.hidden_dim, device=input.device)

        id_list = range(seq_len)

        if not self.left2right:
            id_list = list(reversed(id_list))

        # 用来存储 WordLSTMCell 计算得到的 c^w_t, 就是在某个位置上 word 的 ct
        # 初始是空的，会通过 skip_input， 在计算的过程中逐渐填充
        # 注意: 当 t=0 时, input_c_list[0] 一定是 空的。因为，第一个没有前面的字，所以无法组成词。
        input_c_list = init_list_of_objects(seq_len)

        for t in id_list:

            (hx, cx) = self.rnn(input[t], input_c_list[t], (hx, cx))

            hidden_out.append(hx)
            memory_out.append(cx)

            if skip_input[t]:  # 如果当前 t 位置的字，有词构成，那么，则使用 WordLSTMCell 生成 c^w_t

                # 一共匹配了 多少个 词
                matched_num = len(skip_input[t][0])

                # 获取所有 word id, 组成 word id tensor, 注意是 多个 word id, 不是一个
                word_var = torch.tensor(skip_input[t][0], device=input.device, dtype=torch.long)

                # 获取所有 word id 以及词向量
                word_emb = self.word_emb(word_var)
                word_emb = self.word_dropout(word_emb)

                # 计算所有 word id 的 c^w_t, 一次性计算得到多个 word id 的 c^w
                ct = self.word_rnn(word_emb, (hx, cx))

                assert (ct.size(0) == len(skip_input[t][1]))

                # 将计算得到的所有 c^w_t 全部存放到 input_c_list 对应的 字 的位置上。
                for idx in range(matched_num):

                    length = skip_input[t][1][idx]

                    if self.left2right:
                        # if t+length <= seq_len -1:
                        input_c_list[t + length - 1].append(ct[idx, :].unsqueeze(0))
                    else:
                        # if t-length >=0:
                        input_c_list[t - length + 1].append(ct[idx, :].unsqueeze(0))

        if not self.left2right:
            hidden_out = list(reversed(hidden_out))
            memory_out = list(reversed(memory_out))

        output_hidden, output_memory = torch.cat(hidden_out, 0), torch.cat(memory_out, 0)

        # 输出的 shape: (batch, seq_len, hidden_dim)
        return output_hidden.unsqueeze(0), output_memory.unsqueeze(0)


def init_list_of_objects(size) -> List[List]:
    """
    构建一个 word ct 的 list
    :param size: sequence 的 长度
    :return: 二维 list, 因为每一个 字 的位置，可能有多个词，所以有多个 wrod ct
    """

    list_of_objects = list()
    for i in range(0, size):
        list_of_objects.append(list())

    return list_of_objects


def convert_forward_gaz_to_backward(forward_gaz):
    # print forward_gaz
    length = len(forward_gaz)
    backward_gaz = init_list_of_objects(length)
    for idx in range(length):
        if forward_gaz[idx]:
            assert (len(forward_gaz[idx]) == 2)
            num = len(forward_gaz[idx][0])
            for idy in range(num):
                the_id = forward_gaz[idx][0][idy]
                the_length = forward_gaz[idx][1][idy]
                new_pos = idx + the_length - 1
                if backward_gaz[new_pos]:
                    backward_gaz[new_pos][0].append(the_id)
                    backward_gaz[new_pos][1].append(the_length)
                else:
                    backward_gaz[new_pos] = [[the_id], [the_length]]
    return backward_gaz




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
多 GPU 相关工具

Authors: PanXu
Date:    2020/12/16 16:51:00
"""

from .sync_util import Sync
from .distributed_util import DistributedFuncWrapper


#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2021 PanXu, Inc. All Rights Reserved
#
"""
测试 gat

Authors: PanXu
Date:    2021/02/15 11:19:00
"""

import torch

from easytext.modules import GraphAttentionLayer, GAT
from easytext.utils.nn import tensor_util
from easytext.tests import ASSERT


def test_graph_attention_layer():
    torch.manual_seed(7)
    torch.cuda.manual_seed_all(7)

    in_features = 2
    out_features = 4

    gat_layer = GraphAttentionLayer(in_features=in_features, out_features=out_features, dropout=0.0, alpha=0.1)

    nodes = torch.tensor([[[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]],
                          [[0.7, 0.8], [0.9, 0.10], [0.11, 0.12]]],
                         dtype=torch.float)

    adj = torch.tensor([[[0, 1, 0],
                         [1, 0, 0],
                         [0, 0, 0]],
                        [[0, 1, 1],
                         [1, 0, 1],
                         [1, 1, 0]]],
                       dtype=torch.long)

    outputs: torch.Tensor = gat_layer(input=nodes, adj=adj)

    expect_size = (nodes.size(0), nodes.size(1), out_features)

    ASSERT.assertEqual(expect_size, outputs.size())

    # 下面的 expect 是从原论文中测试得到的结果，直接拿来用
    expect = torch.tensor([[[0.2831, 0.3588, -0.5131, -0.2058],
                            [0.1606, 0.1292, -0.2264, -0.0951],
                            [0.2831, 0.3588, -0.5131, -0.2058]],

                           [[-0.0748, 0.5025, -0.3840, -0.1192],
                            [0.2959, 0.4624, -0.6123, -0.2405],
                            [0.1505, 0.8668, -0.8609, -0.3059]]], dtype=torch.float)

    ASSERT.assertTrue(tensor_util.is_tensor_equal(expect, outputs, epsilon=1e-4))


def test_gat_without_hidden():
    """
    测试 gat
    :return:
    """

    torch.manual_seed(7)
    torch.cuda.manual_seed_all(7)

    in_features = 2
    out_features = 4

    gat = GAT(in_features=in_features,
              out_features=out_features,
              dropout=0.,
              alpha=0.1,
              num_heads=3,
              hidden_size=None)

    nodes = torch.tensor([[[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]],
                          [[0.7, 0.8], [0.9, 0.10], [0.11, 0.12]]],
                         dtype=torch.float)

    adj = torch.tensor([[[0, 1, 0],
                         [1, 0, 0],
                         [0, 0, 0]],
                        [[0, 1, 1],
                         [1, 0, 1],
                         [1, 1, 0]]],
                       dtype=torch.long)

    output_nodes: torch.Tensor = gat(nodes=nodes, adj=adj)

    expect_size = (nodes.size(0), nodes.size(1), out_features)
    ASSERT.assertEqual(expect_size, output_nodes.size())

    expect = torch.tensor([[[-1.6478, -0.3935, -2.6613, -2.7653],
                            [-1.3204, -0.8394, -1.8519, -1.9375],
                            [-1.6478, -0.3935, -2.6613, -2.7653]],

                           [[-1.9897, -0.4203, -2.4447, -2.1232],
                            [-2.1944, -0.1897, -3.4053, -3.5697],
                            [-2.9364, -0.0878, -4.1695, -4.1617]]], dtype=torch.float)

    ASSERT.assertTrue(tensor_util.is_tensor_equal(expect, output_nodes, epsilon=1e-4))


def test_gat_with_hidden():
    """
    测试 gat
    :return:
    """

    torch.manual_seed(7)
    torch.cuda.manual_seed_all(7)

    in_features = 2
    out_features = 4

    gat = GAT(in_features=in_features,
              out_features=out_features,
              dropout=0.,
              alpha=0.1,
              num_heads=3,
              hidden_size=3)

    nodes = torch.tensor([[[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]],
                          [[0.7, 0.8], [0.9, 0.10], [0.11, 0.12]]],
                         dtype=torch.float)

    adj = torch.tensor([[[0, 1, 0],
                         [1, 0, 0],
                         [0, 0, 0]],
                        [[0, 1, 1],
                         [1, 0, 1],
                         [1, 1, 0]]],
                       dtype=torch.long)

    output_nodes: torch.Tensor = gat(nodes=nodes, adj=adj)

    expect_size = (nodes.size(0), nodes.size(1), out_features)
    ASSERT.assertEqual(expect_size, output_nodes.size())

    expect = torch.tensor([[[-1.3835, -1.4764, -1.2033, -1.5113],
                            [-1.3316, -1.5785, -1.1564, -1.5368],
                            [-1.3475, -1.5467, -1.1706, -1.5279]],

                           [[-1.3388, -1.6693, -1.4427, -1.1610],
                            [-1.4288, -1.6525, -1.6607, -0.9707],
                            [-1.4320, -1.4422, -1.6465, -1.1025]]])

    ASSERT.assertTrue(tensor_util.is_tensor_equal(expect, output_nodes, epsilon=1e-4))

